import nengo
import numpy as np
import pytest


@pytest.mark.parametrize('noise_model', ('none', 'alpharc'))
def test_softlif_layer(noise_model, plt):
    pytest.importorskip('keras')
    import keras.models
    import nengo_extras.keras

    nx = 256
    ni = 500

    params = dict(sigma=0.02, amplitude=0.063, tau_rc=0.05, tau_ref=0.001)
    layer_params = dict(params)
    layer_params['noise_model'] = noise_model
    if noise_model == 'alpharc':
        layer_params['tau_s'] = 0.001

    model = keras.models.Sequential()
    model.add(nengo_extras.keras.SoftLIF(input_shape=(nx,), **layer_params))

    x = np.linspace(-10, 30, nx)
    y = model.predict(np.ones((ni, 1)) * x)
    y0 = nengo_extras.neurons.SoftLIFRate(**params).rates(x, 1., 1.)

    plt.plot(x, y.mean(axis=0), 'b')
    if noise_model != 'none':
        plt.plot(x, np.percentile(y, 5, axis=0), 'b:')
        plt.plot(x, np.percentile(y, 95, axis=0), 'b:')
    plt.plot(x, y0, 'k--')

    assert np.isfinite(y).all()
    if noise_model == 'none':
        assert np.allclose(y, y0, atol=1e-3, rtol=1e-3)
    else:
        assert np.allclose(y.mean(axis=0), y0, atol=1e-1, rtol=1e-2)


@pytest.mark.parametrize('noise_model', ('none', 'alpharc'))
def test_softlif_derivative(noise_model, plt):
    pytest.importorskip('keras')
    import keras.models
    import nengo_extras.keras
    from keras import backend as K

    params = dict(sigma=0.02, amplitude=0.063, tau_rc=0.05, tau_ref=0.001)
    layer_params = dict(params)
    layer_params['noise_model'] = noise_model

    model = keras.models.Sequential()
    model.add(nengo_extras.keras.SoftLIF(input_shape=(1,), **layer_params))

    x = model.layers[0].input
    y = model.layers[0].output
    df = K.function([x], K.gradients(K.sum(y), [x]))

    x = np.linspace(-10, 30, 1024).reshape(-1, 1)
    dy = df([x])[0]
    dy0 = nengo_extras.neurons.SoftLIFRate(**params).derivative(x, 1., 1.)

    plt.plot(x, dy)
    plt.plot(x, dy0, 'k--')

    assert np.isfinite(dy).all()
    assert np.allclose(dy, dy0, atol=1e-2, rtol=1e-3)


def test_conv2d_layer(Simulator, seed, rng, plt):
    pytest.importorskip('keras')
    import keras
    import nengo_extras.keras as nekeras
    np.random.seed(seed)  # for Keras weights

    data_format = 'channels_first'
    img_shape = (1, 28, 28)
    presentation_time = 0.01

    nb_filters = 4
    nb_conv = 3

    kmodel = keras.models.Sequential()
    kmodel.add(keras.layers.Conv2D(
        nb_filters, (nb_conv, nb_conv), padding='valid', strides=(2, 2),
        input_shape=img_shape, data_format=data_format))

    X = rng.uniform(-1, 1, size=(1,) + img_shape)
    output_shape = nekeras.kmodel_compute_shapes(kmodel, X.shape)[-1][1:]

    with nengo.Network() as model:
        u = nengo.Node(nengo.processes.PresentInput(X, presentation_time))
        knet = nekeras.SequentialNetwork(
            kmodel, synapse=None, lif_type='lifrate')
        nengo.Connection(u, knet.input, synapse=None)
        output_p = nengo.Probe(knet.output)

    with Simulator(model) as sim:
        sim.run(presentation_time)

    y0 = kmodel.predict(X)[0]
    y1 = sim.data[output_p][-1].reshape(output_shape)
    assert np.allclose(y1, y0, atol=1e-5, rtol=1e-5)



import nengo
from nengo.utils.matplotlib import implot
from nengo.utils.numpy import rms
import numpy as np
import pytest

from nengo_extras.neurons import FastLIF, rates_isi, rates_kernel, SoftLIFRate


def test_softlifrate_rates(plt):
    gain = 0.9
    bias = 1.7
    tau_rc = 0.03
    tau_ref = 0.002

    lif = nengo.LIFRate(tau_rc=tau_rc, tau_ref=tau_ref)
    softlif = SoftLIFRate(tau_rc=tau_rc, tau_ref=tau_ref, sigma=0.00001)

    x = np.linspace(-2, 2, 301)
    lif_r = lif.rates(x, gain, bias)
    softlif_r = softlif.rates(x, gain, bias)

    plt.plot(x, lif_r)
    plt.plot(x, softlif_r)

    assert np.allclose(softlif_r, lif_r, atol=1e-3, rtol=1e-3)


def test_softlifrate_derivative(plt):
    sigma = 0.02
    amplitude = 0.8
    gain = 0.9
    bias = 1.7

    x, dx = np.linspace(-2, 2, 301, retstep=True)
    x2 = 0.5 * (x[:-1] + x[1:])

    neuron = SoftLIFRate(sigma=sigma, amplitude=amplitude)
    r = neuron.rates(x, gain, bias)
    deltar = np.diff(r) / dx
    dr = neuron.derivative(x2, gain, bias)

    plt.subplot(211)
    plt.plot(x, r)

    plt.subplot(212)
    plt.plot(x2, dr)
    plt.plot(x2, deltar, 'k--')

    assert np.allclose(dr, deltar, atol=1e-5, rtol=1e-2)


def test_fastlif(plt):
    """Test that the dynamic model approximately matches the rates."""
    # Based nengo.tests.test_neurons.test_lif
    rng = np.random.RandomState(10)

    dt = 1e-3
    n = 5000
    x = 0.5
    encoders = np.ones((n, 1))
    max_rates = rng.uniform(low=10, high=200, size=n)
    intercepts = rng.uniform(low=-1, high=1, size=n)

    m = nengo.Network()
    with m:
        ins = nengo.Node(x)
        ens = nengo.Ensemble(n, dimensions=1,
                             neuron_type=FastLIF(),
                             encoders=encoders,
                             max_rates=max_rates,
                             intercepts=intercepts)
        nengo.Connection(
            ins, ens.neurons, transform=np.ones((n, 1)), synapse=None)
        spike_probe = nengo.Probe(ens.neurons)
        voltage_probe = nengo.Probe(ens.neurons, 'voltage')
        ref_probe = nengo.Probe(ens.neurons, 'refractory_time')

    t_final = 1.0
    with nengo.Simulator(m, dt=dt) as sim:
        sim.run(t_final)

    i = 3
    plt.subplot(311)
    plt.plot(sim.trange(), sim.data[spike_probe][:, :i])
    plt.subplot(312)
    plt.plot(sim.trange(), sim.data[voltage_probe][:, :i])
    plt.subplot(313)
    plt.plot(sim.trange(), sim.data[ref_probe][:, :i])
    plt.ylim([-dt, ens.neuron_type.tau_ref + dt])

    # check rates against analytic rates
    math_rates = ens.neuron_type.rates(
        x, *ens.neuron_type.gain_bias(max_rates, intercepts))
    spikes = sim.data[spike_probe]
    sim_rates = (spikes > 0).sum(0) / t_final
    print("ME = %f" % (sim_rates - math_rates).mean())
    print("RMSE = %f" % (
        rms(sim_rates - math_rates) / (rms(math_rates) + 1e-20)))
    assert np.sum(math_rates > 0) > 0.5 * n, (
        "At least 50% of neurons must fire")
    assert np.allclose(sim_rates, math_rates, atol=1, rtol=0.02)

    # if voltage and ref time are non-constant, the probe is doing something
    assert np.abs(np.diff(sim.data[voltage_probe])).sum() > 1
    assert np.abs(np.diff(sim.data[ref_probe])).sum() > 1

    # compute spike counts after each timestep
    actual_counts = (spikes > 0).cumsum(axis=0)
    expected_counts = np.outer(sim.trange(), math_rates)
    assert (abs(actual_counts - expected_counts) < 1).all()


def _test_rates(Simulator, rates, plt, seed):
    n = 100
    intercepts = np.linspace(-0.99, 0.99, n)

    model = nengo.Network(seed=seed)
    with model:
        model.config[nengo.Ensemble].max_rates = nengo.dists.Choice([50])
        model.config[nengo.Ensemble].encoders = nengo.dists.Choice([[1]])
        u = nengo.Node(output=nengo.processes.WhiteSignal(2, high=5))
        a = nengo.Ensemble(n, 1,
                           intercepts=intercepts, neuron_type=nengo.LIFRate())
        b = nengo.Ensemble(n, 1,
                           intercepts=intercepts, neuron_type=nengo.LIF())
        nengo.Connection(u, a, synapse=0)
        nengo.Connection(u, b, synapse=0)
        up = nengo.Probe(u)
        ap = nengo.Probe(a.neurons)
        bp = nengo.Probe(b.neurons)

    with Simulator(model, seed=seed+1) as sim:
        sim.run(2.)

    t = sim.trange()
    x = sim.data[up]
    a_rates = sim.data[ap]
    spikes = sim.data[bp]
    b_rates = rates(t, spikes)

    if plt is not None:
        ax = plt.subplot(411)
        plt.plot(t, x)
        ax = plt.subplot(412)
        implot(plt, t, intercepts, a_rates.T, ax=ax)
        ax.set_ylabel('intercept')
        ax = plt.subplot(413)
        implot(plt, t, intercepts, b_rates.T, ax=ax)
        ax.set_ylabel('intercept')
        ax = plt.subplot(414)
        implot(plt, t, intercepts, (b_rates - a_rates).T, ax=ax)
        ax.set_xlabel('time [s]')
        ax.set_ylabel('intercept')

    tmask = (t > 0.1) & (t < 1.9)
    relative_rmse = rms(b_rates[tmask] - a_rates[tmask]) / rms(a_rates[tmask])
    return relative_rmse


def test_rates_isi(Simulator, plt, seed):
    pytest.importorskip('scipy')
    rel_rmse = _test_rates(Simulator, rates_isi, plt, seed)
    assert rel_rmse < 0.3


def test_rates_kernel(Simulator, plt, seed):
    rel_rmse = _test_rates(Simulator, rates_kernel, plt, seed)
    assert rel_rmse < 0.25


@pytest.mark.noassertions
def test_rates(Simulator, seed, logger):
    pytest.importorskip('scipy')
    functions = [
        ('isi_zero', lambda t, s: rates_isi(
            t, s, midpoint=False, interp='zero')),
        ('isi_midzero', lambda t, s: rates_isi(
            t, s, midpoint=True, interp='zero')),
        ('isi_linear', lambda t, s: rates_isi(
            t, s, midpoint=False, interp='linear')),
        ('isi_midlinear', lambda t, s: rates_isi(
            t, s, midpoint=True, interp='linear')),
        ('kernel_expon', lambda t, s: rates_kernel(t, s, kind='expon')),
        ('kernel_gauss', lambda t, s: rates_kernel(t, s, kind='gauss')),
        ('kernel_expogauss', lambda t, s: rates_kernel(
            t, s, kind='expogauss')),
        ('kernel_alpha', lambda t, s: rates_kernel(t, s, kind='alpha')),
    ]

    for name, function in functions:
        rel_rmse = _test_rates(Simulator, function, None, seed)
        logger.info('rate estimator: %s', name)
        logger.info('relative RMSE: %0.4f', rel_rmse)



import nengo
from nengo.exceptions import ValidationError
from nengo.utils.numpy import rms
import numpy as np
import pytest


from nengo_extras.learning_rules import AML, DeltaRule


@pytest.mark.slow
def test_aml(Simulator, seed, rng, plt):
    d = 32
    vocab = nengo.spa.Vocabulary(d, rng=rng)
    n_items = 3
    item_duration = 1.

    def err_stimulus(t):
        if t <= n_items * item_duration:
            v = vocab.parse('Out' + str(int(t // item_duration))).v
        else:
            v = np.zeros(d)
        return np.concatenate(((1., 1.), v))

    def pre_stimulus(t):
        return vocab.parse('In' + str(int((t // item_duration) % n_items))).v

    with nengo.Network(seed=seed) as model:
        pre = nengo.Ensemble(50 * d, d)
        post = nengo.Node(size_in=d)
        c = nengo.Connection(
            pre, post, learning_rule_type=AML(d),
            function=lambda x: np.zeros(d))
        err = nengo.Node(err_stimulus)
        inp = nengo.Node(pre_stimulus)
        nengo.Connection(inp, pre)
        nengo.Connection(err, c.learning_rule)
        p_pre = nengo.Probe(pre, synapse=0.01)
        p_post = nengo.Probe(post, synapse=0.01)
        p_err = nengo.Probe(err, synapse=0.01)

    with Simulator(model) as sim:
        sim.run(2 * n_items * item_duration)

    vocab_out = vocab.create_subset(['Out' + str(i) for i in range(n_items)])
    vocab_in = vocab.create_subset(['In' + str(i) for i in range(n_items)])

    fig = plt.figure()

    ax1 = fig.add_subplot(3, 1, 1)
    ax1.plot(sim.trange(), nengo.spa.similarity(sim.data[p_pre], vocab_in))
    ax1.set_ylabel(r"Cue $\mathbf{u}(t)$")

    ax2 = fig.add_subplot(3, 1, 2, sharex=ax1, sharey=ax1)
    ax2.plot(sim.trange(), nengo.spa.similarity(
        sim.data[p_err][:, 2:], vocab_out))
    ax2.set_ylabel(r"Target $\mathbf{v}(t)$")

    ax3 = fig.add_subplot(3, 1, 3, sharex=ax1, sharey=ax1)
    ax3.plot(sim.trange(), nengo.spa.similarity(sim.data[p_post], vocab_out))
    ax3.set_ylabel("AML output")

    ax1.set_ylim(bottom=0.)

    for ax in [ax1, ax2, ax3]:
        ax.label_outer()
    fig.tight_layout()

    t = sim.trange()
    similarity = nengo.spa.similarity(sim.data[p_post], vocab_out)
    for i in range(n_items):
        assert item_duration > 0.3
        start = (n_items + i) * item_duration + 0.3
        end = (n_items + i + 1) * item_duration
        assert np.all(similarity[(start < t) & (t <= end), i] > 0.8)


@pytest.mark.parametrize('post_target', [None, 'in', 'out'])
def test_delta_rule(Simulator, seed, rng, plt, post_target):
    f = np.cos

    learning_rate = 2e-2

    tau_s = 0.005

    t_train = 10  # amount of learning time
    t_test = 3  # amount of testing time

    n = 50
    max_rate = 200
    dmean = 2. / (n * max_rate)  # theoretical mean for on/off decoders
    dr = 2 * 2 * dmean  # twice mean with additional 2x fudge factor
    decoders = rng.uniform(-dr, dr, size=(1, n))

    ens_params = dict(neuron_type=nengo.LIF(),
                      max_rates=nengo.dists.Choice([max_rate]),
                      intercepts=nengo.dists.Uniform(-1, 0.8))

    if post_target == 'in':
        step = lambda j: (j > 1).astype(j.dtype)
        learning_rule_type = DeltaRule(
            learning_rate=learning_rate, post_fn=step, post_target=post_target,
            post_tau=0.005)
    elif post_target == 'out':
        step = lambda s: (s > 18).astype(s.dtype)
        learning_rule_type = DeltaRule(
            learning_rate=learning_rate, post_fn=step, post_target=post_target,
            post_tau=0.005)
    else:
        learning_rule_type = DeltaRule(learning_rate=learning_rate)

    with nengo.Network(seed=seed) as model:
        u = nengo.Node(nengo.processes.WhiteSignal(period=10, high=5))
        a = nengo.Ensemble(n, 1, **ens_params)
        b = nengo.Ensemble(n, 1, **ens_params)
        y = nengo.Node(size_in=1)

        nengo.Connection(u, a, synapse=None)
        nengo.Connection(b.neurons, y, transform=decoders, synapse=tau_s)
        c = nengo.Connection(a.neurons, b.neurons, synapse=tau_s)

        e = nengo.Node(lambda t, x: x if t < t_train else 0, size_in=1)
        eb = nengo.Node(size_in=n)

        nengo.Connection(u, e, transform=-1, function=f,
                         synapse=nengo.synapses.Alpha(tau_s))
        nengo.Connection(b.neurons, e, transform=decoders, synapse=tau_s)
        nengo.Connection(e, eb, synapse=None, transform=decoders.T)

        c.transform = np.zeros((n, n))
        c.learning_rule_type = learning_rule_type
        nengo.Connection(eb, c.learning_rule, synapse=None)

        ep = nengo.Probe(e)
        up = nengo.Probe(u, synapse=nengo.synapses.Alpha(tau_s))
        yp = nengo.Probe(y)

    with Simulator(model, seed=seed+1) as sim:
        sim.run(t_train + t_test)

    t = sim.trange()
    m = t > t_train
    filt = nengo.synapses.Alpha(0.005)
    x = filt.filtfilt(sim.data[up])
    fx = f(x)
    y = filt.filtfilt(sim.data[yp])

    plt.subplot(311)
    plt.plot(t, sim.data[ep])
    plt.ylabel('error')

    plt.subplot(312)
    plt.plot(t, fx)
    plt.plot(t, y)
    plt.ylabel('output')

    plt.subplot(313)
    plt.plot(t[m], fx[m])
    plt.plot(t[m], y[m])
    plt.ylabel('test output')

    plt.tight_layout()

    rms_error = rms(y[m] - fx[m]) / rms(fx[m])
    assert rms_error < 0.3


def test_delta_rule_function_param_size():
    fn = lambda j: j[:-1]
    with pytest.raises(ValidationError):
        DeltaRule(post_fn=fn)



from __future__ import absolute_import

import matplotlib.pyplot as plt
import numpy as np


def imshow(image, ax=None, vmin=None, vmax=None, invert=False,
           interpolation='none', axes=False):
    """Nicer version of Matplotlib's imshow.

    - By default, show the raw image with no interpolation.
    - If the image is greyscale, use grey colormap.
    """
    kwargs = dict(vmin=vmin, vmax=vmax, interpolation=interpolation)

    if image.ndim == 2 or image.ndim == 3 and image.shape[2] == 1:
        image = image[:, :, 0] if image.ndim == 3 else image
        kwargs['cmap'] = 'gray' if not invert else 'gist_yarg'
    elif image.ndim == 3:
        assert image.shape[2] == 3
        if vmin is not None and vmax is not None:
            image = (image.clip(vmin, vmax) - vmin) / (vmax - vmin)
    else:
        raise ValueError("Wrong number of image dimensions")

    ax = plt.gca() if ax is None else ax
    ax_img = ax.imshow(image, **kwargs)
    if not axes:
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)

    return ax_img


def tile(images, ax=None, rows=9, cols=12,
         grid=False, gridwidth=1, gridcolor='r', **show_params):
    """Plot a grid of images

    Parameters
    ----------
    images : ndarray (n_images, height, width, channels)
        Array of images to display.
    """
    if images.ndim == 3:
        images = images[:, :, :, None]
    n_images, m, n, nc = images.shape

    if n_images < rows * cols:
        aspect = float(cols) / rows
        rows = int(np.maximum(np.round(np.sqrt(n_images / aspect)), 1))
        cols = int(np.ceil(float(n_images) / rows))

    img_shape = (m*rows, n*cols, nc)
    img = np.zeros(img_shape, dtype=images.dtype)
    for k in range(min(rows*cols, n_images)):
        i, j = k // cols, k % cols
        img[i*m:(i+1)*m, j*n:(j+1)*n, :] = images[k]

    ax_img = imshow(img, ax=ax, **show_params)
    ax = ax_img.axes

    if grid:
        for i in range(1, rows):
            ax.plot([-0.5, img.shape[1]-0.5], [i*m-0.5, i*m-0.5],
                    color=gridcolor, linewidth=gridwidth)
        for j in range(1, cols):
            ax.plot([j*n-0.5, j*n-0.5], [-0.5, img.shape[0]-0.5],
                    color=gridcolor, linewidth=gridwidth)

        ax.set_xlim([-0.5, img.shape[1]-0.5])
        ax.set_ylim([-0.5, img.shape[0]-0.5])
        ax.invert_yaxis()


def compare(image_sets, ax=None, rows=4, cols=12,
            grid=True, gridwidth=1, gridcolor='r', **show_params):
    """Compare sets of images in a grid.

    Parameters
    ----------
    image_sets : list of (n_images, height, width, channels) ndarray
        List of the sets of images to compare. Each set of images must be
        the same size.
    """
    d = len(image_sets)

    n_images = image_sets[0].shape[0]
    imshape = image_sets[0].shape[1:]
    m, n = imshape[:2]
    nc = imshape[2] if len(imshape) > 2 else 1
    for q in range(d):
        if image_sets[q].shape != image_sets[0].shape:
            raise ValueError(
                "All image sets must be the same shape as the first")

    if n_images < rows * cols:
        aspect = float(cols) / rows
        rows = int(np.maximum(np.round(np.sqrt(n_images / aspect)), 1))
        cols = int(np.ceil(float(n_images) / rows))

    img_shape = (d*m*rows, n*cols, nc)
    img = np.zeros(img_shape, dtype=image_sets[0].dtype)

    for k in range(min(rows*cols, n_images)):
        i, j = k // cols, k % cols
        for q in range(d):
            img[(d*i+q)*m:(d*i+q+1)*m, j*n:(j+1)*n] = \
                image_sets[q][k, :].reshape(m, n, nc)

    ax_img = imshow(img, ax=ax, **show_params)
    ax = ax_img.axes

    if grid:
        for i in range(1, rows):
            ax.plot([-0.5, img.shape[1]-0.5], (d*i*m-0.5)*np.ones(2),
                    color=gridcolor, linewidth=gridwidth)
        for j in range(1, cols):
            ax.plot([j*n-0.5, j*n-0.5], [-0.5, img.shape[0]-0.5],
                    color=gridcolor, linewidth=gridwidth)

        ax.set_xlim([-0.5, img.shape[1]-0.5])
        ax.set_ylim([-0.5, img.shape[0]-0.5])
        ax.invert_yaxis()



"""
Tools for visualizing deep neural networks
"""
import collections

import numpy as np
import PIL.ImageTk

from .compat import tkinter as tk


class ImageSelector(tk.Frame):
    """Choose between images with left and right arrows

    Attributes
    ----------
    image_function : callable
        Turn an array into a PIL.Image. Can be generated by
        ``nengo_extras.gui.image_function``.
    resample : int
        Resampling mode for ``PIL.Image.resize``.
    """

    def __init__(self, parent, **kwargs):
        tk.Frame.__init__(self, parent, **kwargs)

        self.image_function = None
        self.resample = 0

        self.canvas = tk.Canvas(self)
        self.canvas_image = self.canvas.create_image(0, 0, anchor='nw')
        self.left = tk.Button(self, text='<')
        self.right = tk.Button(self, text='>')

        self.canvas.bind('<Configure>', self._on_canvas_resize)

        # --- layout
        self.left.pack(side='left', expand=False)
        self.right.pack(side='right', expand=False)
        self.canvas.pack(fill='both', expand=True)

    def set_left_callback(self, callback):
        self.left.configure(command=callback)

    def set_right_callback(self, callback):
        self.right.configure(command=callback)

    def _canvas_resized(self):
        h = self.canvas.winfo_height()
        w = self.canvas.winfo_width()
        image = self.image.resize((w, h), self.resample)
        photo = PIL.ImageTk.PhotoImage(image)
        self.photo = photo
        self.canvas.itemconfig(self.canvas_image, image=photo)

    def _on_canvas_resize(self, event):
        self._canvas_resized()

    def set_image(self, image_array):
        self.image = self.image_function(image_array)
        self._canvas_resized()


class ScrollCanvasFrame(tk.Frame):
    def __init__(self, parent, vertical=False, horizontal=False, **kwargs):
        tk.Frame.__init__(self, parent, **kwargs)

        self.vertical = vertical
        self.horizontal = horizontal

        # self.canvas = tk.Canvas(frame, height=size[0], width=size[1])
        self.canvas = tk.Canvas(self)

        if vertical:
            self.vscrollbar = tk.Scrollbar(
                self, orient='vertical', command=self.canvas.yview)
            self.canvas.configure(yscrollcommand=self.vscrollbar.set)
            self.vscrollbar.pack(side='right', fill='y')

        if horizontal:
            self.hscrollbar = tk.Scrollbar(
                self, orient='horizontal', command=self.canvas.xview)
            self.canvas.configure(xscrollcommand=self.hscrollbar.set)
            self.hscrollbar.pack(side='bottom', fill='both')

        self.canvas.pack(fill='both', expand=True)
        self.bind('<Configure>', self._on_resize)

    def _on_resize(self, event):
        self.canvas.configure(scrollregion=self.canvas.bbox('all'))


class ScrollWindow(ScrollCanvasFrame):
    """Scroll a single window"""

    def __init__(self, *args, **kwargs):
        ScrollCanvasFrame.__init__(self, *args, **kwargs)
        self.item = self.canvas.create_window((0, 0), anchor='nw')
        self.canvas.bind('<Configure>', self._on_canvas_resize)

    def set_window(self, window):
        self.canvas.itemconfig(self.item, window=window)

    def _on_canvas_resize(self, event):
        if not self.vertical and not self.horizontal:
            h = self.canvas.winfo_height()
            w = self.canvas.winfo_width()
            self.canvas.itemconfig(self.item, height=h, width=w)
        elif not self.vertical:
            h = self.canvas.winfo_height()
            self.canvas.itemconfig(self.item, height=h)
        elif not self.horizontal:
            w = self.canvas.winfo_width()
            self.canvas.itemconfig(self.pane_item, width=w)


class VerticalImageFrame(ScrollCanvasFrame):

    Column = collections.namedtuple(
        'Column', ('width', 'height', 'image_function'))

    def __init__(self, parent, **kwargs):
        ScrollCanvasFrame.__init__(
            self, parent, vertical=True, **kwargs)

        self.padding = (5, 5)  # width, height

        self.columns = []
        self.canvas_images = []  # by row, then column
        self.images = []  # by row, then column
        self.photos = []  # by row, then column

    @property
    def n_columns(self):
        return len(self.columns)

    @property
    def n_rows(self):
        return len(self.canvas_images)

    @property
    def column_heights(self):
        return (column.height for column in self.columns)

    @property
    def column_widths(self):
        return (column.width for column in self.columns)

    @property
    def column_image_functions(self):
        return (column.image_function for column in self.columns)

    def add_column(self, width, height, image_function):
        self.columns.append(self.Column(width, height, image_function))
        self.canvas.configure(width=sum(self.column_widths))

    def create_canvas_images(self, n):
        for image_row in self.canvas_images:
            for image in image_row:
                self.canvas.delete(image)
        del self.canvas_images[:]

        max_height = max(self.column_heights)
        widths = np.array(list(self.column_widths))
        pos_x = np.cumsum([0] + list(widths + self.padding[0]))[:-1]

        for i in range(n):
            image_row = [
                self.canvas.create_image(
                    pos_x[j], i*(max_height + self.padding[1]), anchor='nw')
                for j in range(self.n_columns)]
            self.canvas_images.append(image_row)

    def set_images(self, column_images):
        assert len(column_images) == self.n_columns
        n_rows = max(len(images) for images in column_images)
        if n_rows > self.n_rows:
            self.create_canvas_images(n_rows)

        del self.images[:]
        del self.photos[:]

        for i in range(self.n_rows):
            images = []
            photos = []
            for j in range(self.n_columns):
                width, height, image_function = self.columns[j]

                f = (column_images[j][i] if i < len(column_images[j]) else
                     np.zeros_like(column_images[j][0]))
                image = image_function(f)

                photo = PIL.ImageTk.PhotoImage(image.resize((width, height)))
                self.canvas.itemconfig(self.canvas_images[i][j], image=photo)

                images.append(image)
                photos.append(photo)

            self.images.append(images)
            self.photos.append(photos)


class Viewer(tk.Tk):
    def __init__(self, images, image_function, *args, **kwargs):
        tk.Tk.__init__(self, *args, **kwargs)

        self.filter_size = (50, 50)  # width, height
        self.act_size = (150, 150)  # width, height

        # --- data
        self.images = images
        self.frame_kind_data = {}

        # --- components
        self.selector = ImageSelector(self)
        self.selector.image_function = image_function
        self.selector.set_left_callback(self._on_prev_image)
        self.selector.set_right_callback(self._on_next_image)

        self.pane_frame = ScrollWindow(self, horizontal=True)
        self.panes = tk.PanedWindow(self.pane_frame, orient='horizontal')
        self.pane_frame.set_window(self.panes)

        self.frames = []

        # --- layout
        self.selector.pack(side='top', expand=False)
        self.pane_frame.pack(fill='both', expand=True)

    @property
    def n_frames(self):
        return len(self.frames)

    def add_column(self, kinds, data, fns, title=None):
        assert len(kinds) == len(data) == len(fns)

        pane = tk.Frame(self.panes)
        if title is not None:
            label = tk.Label(pane, text=title)
            label.pack(side='top', fill='both')

        frame = VerticalImageFrame(pane)
        for kind, fn in zip(kinds, fns):
            if kind == 'acts':
                frame.add_column(self.act_size[0], self.act_size[1], fn)
            elif kind == 'filters':
                frame.add_column(self.filter_size[0], self.filter_size[1], fn)
            else:
                raise ValueError("Unrecognized kind %r" % kind)

        frame.pack(fill='both', expand=True)

        self.frames.append(frame)
        self.frame_kind_data[frame] = tuple(zip(kinds, data))
        self.panes.add(pane)

    def add_acts(self, acts, acts_fn, **kwargs):
        self.add_column(('acts',), (acts,), (acts_fn,), **kwargs)

    def add_filters(self, filters, filters_fn, **kwargs):
        self.add_column(('filters',), (filters,), (filters_fn,), **kwargs)

    def add_filters_acts(self, filters, filters_fn, acts, acts_fn, **kwargs):
        self.add_column(
            ('filters', 'acts'), (filters, acts), (filters_fn, acts_fn),
            **kwargs)

    def set_index(self, i):
        self.i = i % len(self.images)
        self.selector.set_image(self.images[self.i])

        is_static = lambda kind: (kind == 'filters')

        for frame in self.frames:
            images = [d if is_static(k) else d[self.i]
                      for k, d in self.frame_kind_data[frame]]
            frame.set_images(images)

    def _on_prev_image(self):
        self.set_index(self.i - 1)

    def _on_next_image(self):
        self.set_index(self.i + 1)



import numpy as np

import nengo
from nengo.utils.compat import range
from nengo.utils.numpy import rmse

import nengo_extras


def test_sine_waves(Simulator, plt, seed):
    radius = 2
    dim = 5
    product = nengo_extras.networks.Product(
        200, dim, radius, net=nengo.Network(seed=seed))

    func_A = lambda t: np.sqrt(radius)*np.sin(np.arange(1, dim+1)*2*np.pi*t)
    func_B = lambda t: np.sqrt(radius)*np.sin(np.arange(dim, 0, -1)*2*np.pi*t)
    with product:
        input_A = nengo.Node(func_A)
        input_B = nengo.Node(func_B)
        nengo.Connection(input_A, product.A)
        nengo.Connection(input_B, product.B)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    AB = np.asarray(list(map(func_A, t))) * np.asarray(list(map(func_B, t)))
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim+1, 1, i+1)
        plt.plot(t + delay, AB[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rmse(AB[:len(offset), :], sim.data[p][offset, :]) < 0.2


def test_direct_mode_with_single_neuron(Simulator, plt, seed):
    radius = 2
    dim = 5

    config = nengo.Config(nengo.Ensemble)
    config[nengo.Ensemble].neuron_type = nengo.Direct()
    with config:
        product = nengo_extras.networks.Product(
            1, dim, radius, net=nengo.Network(seed=seed))

    func_A = lambda t: np.sqrt(radius)*np.sin(np.arange(1, dim+1)*2*np.pi*t)
    func_B = lambda t: np.sqrt(radius)*np.sin(np.arange(dim, 0, -1)*2*np.pi*t)
    with product:
        input_A = nengo.Node(func_A)
        input_B = nengo.Node(func_B)
        nengo.Connection(input_A, product.A)
        nengo.Connection(input_B, product.B)
        p = nengo.Probe(product.output, synapse=0.005)

    with Simulator(product) as sim:
        sim.run(1.0)

    t = sim.trange()
    AB = np.asarray(list(map(func_A, t))) * np.asarray(list(map(func_B, t)))
    delay = 0.013
    offset = np.where(t >= delay)[0]

    for i in range(dim):
        plt.subplot(dim+1, 1, i+1)
        plt.plot(t + delay, AB[:, i])
        plt.plot(t, sim.data[p][:, i])
        plt.xlim(right=t[-1])
        plt.yticks((-2, 0, 2))

    assert rmse(AB[:len(offset), :], sim.data[p][offset, :]) < 0.2



# -*- coding: utf-8 -*-
#
# This file is execfile()d with the current directory set
# to its containing dir.

import sys

try:
    import nengo_extras
    import guzzle_sphinx_theme
except ImportError:
    print("To build the documentation, nengo_extras and guzzle_sphinx_theme "
          "must be installed in the current environment. Please install these "
          "and their requirements first. A virtualenv is recommended!")
    sys.exit(1)

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.githubpages',
    'sphinx.ext.intersphinx',
    'sphinx.ext.mathjax',
    'sphinx.ext.todo',
    'sphinx.ext.viewcode',
    'guzzle_sphinx_theme',
    'numpydoc',
    'nbsphinx',
    'nbsphinx_link',
]

default_role = 'py:obj'

# -- sphinx.ext.autodoc
autoclass_content = 'both'  # class and __init__ docstrings are concatenated
autodoc_default_flags = ['members']
autodoc_member_order = 'bysource'  # default is alphabetical

# -- sphinx.ext.intersphinx
intersphinx_mapping = {
    'nengo': ('https://www.nengo.ai/nengo/', None),
    'numpy': ('https://docs.scipy.org/doc/numpy', None),
    'scipy': ('https://docs.scipy.org/doc/scipy/reference', None),
}

# -- sphinx.ext.todo
todo_include_todos = True

# -- numpydoc
numpydoc_show_class_members = False

# -- nbsphinx
nbsphinx_timeout = -1

# -- sphinx
nitpicky = True
exclude_patterns = ['_build']
source_suffix = '.rst'
source_encoding = 'utf-8'
master_doc = 'index'

# Need to include https Mathjax path for sphinx < v1.3
mathjax_path = ("https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.3/"
                "config/TeX-AMS-MML_HTMLorMML.js")

project = u'Nengo Extras'
authors = u'Applied Brain Research'
copyright = nengo_extras.__copyright__
version = '.'.join(nengo_extras.__version__.split('.')[:2])
release = nengo_extras.__version__  # Full version, with tags
pygments_style = 'default'

# -- Options for HTML output --------------------------------------------------

pygments_style = "sphinx"
templates_path = ["_templates"]
html_static_path = ["_static"]

html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_theme = "guzzle_sphinx_theme"

html_theme_options = {
    "project_nav_name": "Nengo extras %s" % (version,),
    "base_url": "https://www.nengo.ai/nengo-extras",
}

html_title = "Nengo extras {0} docs".format(release)
htmlhelp_basename = 'Nengo extras'
html_last_updated_fmt = ''  # Suppress 'Last updated on:' timestamp
html_show_sphinx = False

# -- Options for LaTeX output -------------------------------------------------

latex_elements = {
    'papersize': 'letterpaper',
    'pointsize': '11pt',
    # 'preamble': '',
}

latex_documents = [
    # (source start file, target, title, author, documentclass [howto/manual])
    ('index', 'nengo_extras.tex', html_title, authors, 'manual'),
]

# -- Options for manual page output -------------------------------------------

man_pages = [
    # (source start file, name, description, authors, manual section).
    ('index', 'nengo_extras', html_title, [authors], 1)
]

# -- Options for Texinfo output -----------------------------------------------

texinfo_documents = [
    # (source start file, target, title, author, dir menu entry,
    #  description, category)
    ('index', 'nengo_extras', html_title, authors, 'Nengo',
     'Lesser used features for Nengo', 'Miscellaneous'),
]



import matplotlib.patches
import nengo
import numpy as np
import pytest

from nengo_extras.plot_spikes import (
    cluster, merge, plot_spikes, sample_by_activity, sample_by_variance)


@pytest.mark.noassertions
def test_plot_spikes(plt, seed, RefSimulator):
    with nengo.Network(seed=seed) as model:
        ens = nengo.Ensemble(10, 1)
        inp = nengo.Node(np.sin)
        nengo.Connection(inp, ens)
        p = nengo.Probe(ens.neurons, 'spikes')

    with RefSimulator(model) as sim:
        sim.run(1.)

    ax = plt.gca()
    ax.add_patch(matplotlib.patches.Rectangle(
        (0, 0), 1, 10,
        fc=(0.8, 0.6, 0.6)))  # To check for transparency
    plot_spikes(sim.trange(), sim.data[p], ax=ax, zorder=1)


def test_cluster():
    dt = 0.001
    t = np.arange(0., 1., dt) + dt

    spikes = np.zeros((len(t), 4))
    spikes[:, 1::2] = 1. / dt

    t_clustered, spikes_clustered = cluster(t, spikes, filter_width=dt)
    assert (t_clustered == t).all()
    assert (spikes_clustered == spikes[:, [0, 2, 1, 3]]).all()


def test_merge():
    dt = 0.001
    t = np.arange(0., 1., dt) + dt

    spikes = np.zeros((len(t), 4))
    spikes[:, 1::2] = 1. / dt
    spikes[:, 0] = 1. / dt

    expected = np.empty((len(t), 2))
    expected[:, 0] = 1. / dt
    expected[:, 1] = .5 / dt

    t_merged, spikes_merged = merge(t, spikes, num=2)
    assert (t_merged == t).all()
    assert (spikes_merged == expected).all()


def test_sample_by_variance():
    dt = 0.001
    t = np.arange(0., 1., dt) + dt

    spikes = np.zeros((len(t), 4))
    spikes[::, 1] = 1. / dt
    spikes[::10, 2] = 1. / dt
    spikes[::100, 3] = 1. / dt

    t_sampled, spikes_sampled = sample_by_variance(
        t, spikes, num=1, filter_width=0.001)
    assert (t_sampled == t).all()
    assert (spikes_sampled == spikes[:, [2]]).all()

    t_sampled, spikes_sampled = sample_by_variance(
        t, spikes, num=1, filter_width=0.1)
    assert (t_sampled == t).all()
    assert (spikes_sampled == spikes[:, [3]]).all()

    t_sampled, spikes_sampled = sample_by_variance(
        t, spikes, num=2, filter_width=0.1)
    assert (t_sampled == t).all()
    assert (spikes_sampled == spikes[:, [3, 2]]).all()

    t_sampled, spikes_sampled = sample_by_variance(
        t, spikes, num=20, filter_width=0.1)
    assert (t_sampled == t).all()
    assert (spikes_sampled == spikes[:, [3, 2, 1, 0]]).all()


def test_sample_by_activity():
    dt = 0.001
    t = np.arange(0., 1., dt) + dt

    spikes = np.zeros((len(t), 4))
    spikes[:, 1::2] = 1. / dt

    t_sampled, spikes_sampled = sample_by_activity(t, spikes, num=2)
    assert (t_sampled == t).all()
    assert (spikes_sampled == np.ones((len(t), 2)) / dt).all()

    t_sampled, spikes_sampled = sample_by_activity(
        t, spikes, num=1, blocksize=2)
    assert (t_sampled == t).all()
    assert (spikes_sampled == np.ones((len(t), 1)) / dt).all()

    t_sampled, spikes_sampled = sample_by_activity(
        t, spikes, num=2, blocksize=2)
    assert (t_sampled == t).all()
    assert (spikes_sampled == np.ones((len(t), 2)) / dt).all()



import numpy as np
import pytest


from nengo_extras.deepnetworks import ConvLayer, LocalLayer, NeuronLayer


@pytest.mark.parametrize('local', (False, True))
def test_convlayer_theano(local, rng):
    pytest.importorskip('theano')
    import theano
    import theano.tensor as tt

    n = 2
    nf = 3
    nc = 4
    nxi, nxj = 6, 8
    si, sj = 3, 5
    pi, pj = 2, 1
    sti, stj = 1, 1
    nyi = 1 + max(int(np.ceil(float(2*pi + nxi - si) / sti)), 0)
    nyj = 1 + max(int(np.ceil(float(2*pj + nxj - sj) / stj)), 0)

    if local:
        filters = rng.uniform(-1, 1, size=(nf, nyi, nyj, nc, si, sj))
    else:
        filters = rng.uniform(-1, 1, size=(nf, nc, si, sj))

    biases = rng.uniform(-1, 1, size=(nf,))
    inputs = rng.uniform(-1, 1, size=(n, nc, nxi, nxj))

    if local:
        layer = LocalLayer((nc, nxi, nxj), filters, biases,
                           strides=(sti, stj), padding=(pi, pj))
    else:
        layer = ConvLayer((nc, nxi, nxj), filters, biases,
                          strides=(sti, stj), padding=(pi, pj))

    y0 = layer.compute(inputs.reshape(n, -1))

    sx = tt.matrix()
    f = theano.function([sx], layer.theano(sx), allow_input_downcast=True)
    y1 = f(inputs.reshape(n, -1))

    assert y0.shape == y1.shape == (n, nf*nyi*nyj)
    assert np.allclose(y1, y0, atol=1e-7)


@pytest.mark.xfail
def test_neuronlayer_softlif_theano():
    pytest.importorskip('theano')
    pytest.importorskip('keras')
    import theano
    import theano.tensor as tt
    from nengo_extras.keras import SoftLIF as SoftLIFLayer
    from nengo_extras.neurons import SoftLIFRate

    klayer = SoftLIFLayer()
    sx = tt.matrix()
    f0 = theano.function([sx], klayer.call(sx), allow_input_downcast=True)

    neuron_type = SoftLIFRate()
    layer = NeuronLayer(1, neuron_type=neuron_type, bias=1)
    sx = tt.matrix()
    f1 = theano.function([sx], layer.theano(sx), allow_input_downcast=True)

    x = np.linspace(-10, 10, 201).reshape(1, -1)
    y0 = f0(x)
    y1 = f1(x)

    assert y0.shape == y1.shape
    assert np.allclose(y0, y1, atol=1e-7)



import numpy as np

from nengo.processes import Process
from nengo.params import EnumParam, NdarrayParam, NumberParam, ShapeParam
from nengo.utils.compat import is_iterable, range


def softmax(x, axis=None):
    """Stable softmax function"""
    ex = np.exp(x - x.max(axis=axis, keepdims=True))
    return ex / ex.sum(axis=axis, keepdims=True)


class Conv2d(Process):
    """Perform 2-D (image) convolution on an input.

    Parameters
    ----------
    shape_in : 3-tuple (n_channels, height, width)
        Shape of the input images: channels, height, width.
    filters : array_like (n_filters, n_channels, f_height, f_width)
        Static filters to convolve with the input. Shape is number of filters,
        number of input channels, filter height, and filter width. Shape can
        also be (n_filters, height, width, n_channels, f_height, f_width)
        to apply different filters at each point in the image, where 'height'
        and 'width' are the input image height and width.
    biases : array_like (1,) or (n_filters,) or (n_filters, height, width)
        Biases to add to outputs. Can have one bias across the entire output
        space, one bias per filter, or a unique bias for each output pixel.
    strides : 2-tuple (vertical, horizontal) or int
        Spacing between filter placements. If an integer
        is provided, the same spacing is used in both dimensions.
    padding : 2-tuple (vertical, horizontal) or int
        Amount of zero-padding around the outside of the input image. Padding
        is applied to both sides, e.g. ``padding=(1, 0)`` will add one pixel
        of padding to the top and bottom, and none to the left and right.
    """

    shape_in = ShapeParam('shape_in', length=3, low=1)
    shape_out = ShapeParam('shape_out', length=3, low=1)
    strides = ShapeParam('strides', length=2, low=1)
    padding = ShapeParam('padding', length=2)
    filters = NdarrayParam('filters', shape=('...',))
    biases = NdarrayParam('biases', shape=('...',), optional=True)
    border = EnumParam('border', values=('floor', 'ceil'))

    def __init__(self, shape_in, filters, biases=None, strides=1, padding=0,
                 border='ceil'):  # noqa: C901
        self.shape_in = shape_in
        self.filters = filters
        if self.filters.ndim not in [4, 6]:
            raise ValueError(
                "`filters` must have four or six dimensions "
                "(filters, [height, width,] channels, f_height, f_width)")
        if self.filters.shape[-3] != self.shape_in[0]:
            raise ValueError(
                "Filter channels (%d) and input channels (%d) must match"
                % (self.filters.shape[-3], self.shape_in[0]))
        if not all(s % 2 == 1 for s in self.filters.shape[-2:]):
            raise ValueError("Filter shapes must be odd (got %r)"
                             % (self.filters.shape[-2:],))

        self.strides = strides if is_iterable(strides) else [strides] * 2
        self.padding = padding if is_iterable(padding) else [padding] * 2
        self.border = border

        nf = self.filters.shape[0]
        nxi, nxj = self.shape_in[1:]
        si, sj = self.filters.shape[-2:]
        pi, pj = self.padding
        sti, stj = self.strides
        rounder = np.ceil if self.border == 'ceil' else np.floor
        nyi = 1 + max(int(rounder(float(2*pi + nxi - si) / sti)), 0)
        nyj = 1 + max(int(rounder(float(2*pj + nxj - sj) / stj)), 0)
        self.shape_out = (nf, nyi, nyj)
        if self.filters.ndim == 6 and self.filters.shape[1:3] != (nyi, nyj):
            raise ValueError("Number of local filters %r must match out shape "
                             "%r" % (self.filters.shape[1:3], (nyi, nyj)))

        self.biases = biases if biases is not None else None
        if self.biases is not None:
            if self.biases.size == 1:
                self.biases.shape = (1, 1, 1)
            elif self.biases.size == np.prod(self.shape_out):
                self.biases.shape = self.shape_out
            elif self.biases.size == self.shape_out[0]:
                self.biases.shape = (self.shape_out[0], 1, 1)
            elif self.biases.size == np.prod(self.shape_out[1:]):
                self.biases.shape = (1,) + self.shape_out[1:]
            else:
                raise ValueError(
                    "Biases size (%d) does not match output shape %s"
                    % (self.biases.size, self.shape_out))

        super(Conv2d, self).__init__(
            default_size_in=np.prod(self.shape_in),
            default_size_out=np.prod(self.shape_out))

    def make_step(self, shape_in, shape_out, dt, rng):
        assert np.prod(shape_in) == np.prod(self.shape_in)
        assert np.prod(shape_out) == np.prod(self.shape_out)
        shape_in, shape_out = self.shape_in, self.shape_out

        filters = self.filters
        local_filters = filters.ndim == 6
        biases = self.biases

        nc, nxi, nxj = shape_in
        nf, nyi, nyj = shape_out
        si, sj = filters.shape[-2:]
        pi, pj = self.padding
        sti, stj = self.strides

        def step_conv2d(t, x):
            x = x.reshape(-1, nc, nxi, nxj)
            n = x.shape[0]
            y = np.zeros((n, nf, nyi, nyj), dtype=x.dtype)

            for i in range(nyi):
                for j in range(nyj):
                    i0 = i*sti - pi
                    j0 = j*stj - pj
                    i1, j1 = i0 + si, j0 + sj
                    sli = slice(max(-i0, 0), min(nxi + si - i1, si))
                    slj = slice(max(-j0, 0), min(nxj + sj - j1, sj))
                    w = (filters[:, i, j, :, sli, slj] if local_filters else
                         filters[:, :, sli, slj])
                    xij = x[:, :, max(i0, 0):min(i1, nxi),
                            max(j0, 0):min(j1, nxj)]
                    y[:, :, i, j] = np.dot(
                        xij.reshape(n, -1), w.reshape(nf, -1).T)

            if biases is not None:
                y += biases

            return y.ravel()

        return step_conv2d


class Pool2d(Process):
    """Perform 2-D (image) pooling on an input.

    Parameters
    ----------
    shape_in : 3-tuple (channels, height, width)
        Shape of the input image.
    pool_size : 2-tuple (vertical, horizontal) or int
        Shape of the pooling region. If an integer is provided, the shape will
        be square with the given side length.
    strides : 2-tuple (vertical, horizontal) or int
        Spacing between pooling placements. If ``None`` (default), will be
        equal to ``pool_size`` resulting in non-overlapping pooling.
    kind : "avg" or "max"
        Type of pooling to perform: average pooling or max pooling.
    mode : "full" or "valid"
        If the input image does not divide into an integer number of pooling
        regions, whether to add partial pooling regions for the extra
        pixels ("full"), or discard extra input pixels ("valid").

    Attributes
    ----------
    shape_out : 3-tuple (channels, height, width)
        Shape of the output image.
    """
    shape_in = ShapeParam('shape_in', length=3, low=1)
    shape_out = ShapeParam('shape_out', length=3, low=1)
    pool_size = ShapeParam('pool_size', length=2, low=1)
    strides = ShapeParam('strides', length=2, low=1)
    kind = EnumParam('kind', values=('avg', 'max'))
    mode = EnumParam('mode', values=('full', 'valid'))

    def __init__(self, shape_in, pool_size, strides=None,
                 kind='avg', mode='full'):
        self.shape_in = shape_in
        self.pool_size = (pool_size if is_iterable(pool_size) else
                          [pool_size] * 2)
        self.strides = (strides if is_iterable(strides) else
                        [strides] * 2 if strides is not None else
                        self.pool_size)
        self.kind = kind
        self.mode = mode
        if not all(st <= p for st, p in zip(self.strides, self.pool_size)):
            raise ValueError("Strides %s must be <= pool_size %s" %
                             (self.strides, self.pool_size))

        nc, nxi, nxj = self.shape_in
        nyi_float = float(nxi - self.pool_size[0]) / self.strides[0]
        nyj_float = float(nxj - self.pool_size[1]) / self.strides[1]
        if self.mode == 'full':
            nyi = 1 + int(np.ceil(nyi_float))
            nyj = 1 + int(np.ceil(nyj_float))
        elif self.mode == 'valid':
            nyi = 1 + int(np.floor(nyi_float))
            nyj = 1 + int(np.floor(nyj_float))
        self.shape_out = (nc, nyi, nyj)

        super(Pool2d, self).__init__(
            default_size_in=np.prod(self.shape_in),
            default_size_out=np.prod(self.shape_out))

    def make_step(self, shape_in, shape_out, dt, rng):
        assert np.prod(shape_in) == np.prod(self.shape_in)
        assert np.prod(shape_out) == np.prod(self.shape_out)
        nc, nxi, nxj = self.shape_in
        nc, nyi, nyj = self.shape_out
        si, sj = self.pool_size
        sti, stj = self.strides
        kind = self.kind
        nxi2, nxj2 = nyi * sti, nyj * stj

        def step_pool2d(t, x):
            x = x.reshape(-1, nc, nxi, nxj)
            y = np.zeros((x.shape[0], nc, nyi, nyj), dtype=x.dtype)
            n = np.zeros((nyi, nyj))

            for i in range(si):
                for j in range(sj):
                    xij = x[:, :, i:min(nxi2+i, nxi):sti,
                            j:min(nxj2+j, nxj):stj]
                    ni, nj = xij.shape[-2:]
                    if kind == 'max':
                        y[:, :, :ni, :nj] = np.maximum(y[:, :, :ni, :nj], xij)
                    elif kind == 'avg':
                        y[:, :, :ni, :nj] += xij
                        n[:ni, :nj] += 1
                    else:
                        raise NotImplementedError(kind)

            if kind == 'avg':
                y /= n

            return y.ravel()

        return step_pool2d


class PresentJitteredImages(Process):
    images = NdarrayParam('images', shape=('...',))
    image_shape = ShapeParam('image_shape', length=3, low=1)
    output_shape = ShapeParam('output_shape', length=2, low=1)
    presentation_time = NumberParam('presentation_time', low=0, low_open=True)
    jitter_std = NumberParam('jitter_std', low=0, low_open=True, optional=True)
    jitter_tau = NumberParam('jitter_tau', low=0, low_open=True)

    def __init__(self, images, presentation_time, output_shape,
                 jitter_std=None, jitter_tau=None, **kwargs):
        import scipy.ndimage.interpolation
        # ^ required for simulation, so check it here

        self.images = images
        self.presentation_time = presentation_time
        self.image_shape = images.shape[1:]
        self.output_shape = output_shape
        self.jitter_std = jitter_std
        self.jitter_tau = (presentation_time if jitter_tau is None else
                           jitter_tau)

        nc = self.image_shape[0]
        nyi, nyj = self.output_shape
        super(PresentJitteredImages, self).__init__(
            default_size_in=0, default_size_out=nc*nyi*nyj, **kwargs)

    def make_step(self, shape_in, shape_out, dt, rng):
        import scipy.ndimage.interpolation

        nc, nxi, nxj = self.image_shape
        nyi, nyj = self.output_shape
        ni, nj = nxi - nyi, nxj - nyj
        nij = np.array([ni, nj])
        assert shape_in == (0,)
        assert shape_out == (nc*nyi*nyj,)

        if self.jitter_std is None:
            si, sj = ni / 4., nj / 4.
        else:
            si = sj = self.jitter_std

        tau = self.jitter_tau

        n = len(self.images)
        images = self.images.reshape((n, nc, nxi, nxj))
        presentation_time = float(self.presentation_time)

        cij = (nij - 1) / 2.
        dt7tau = dt / tau
        sigma2 = np.sqrt(2.*dt/tau) * np.array([si, sj])
        ij = cij.copy()

        def step_presentjitteredimages(t):
            # update jitter position
            ij0 = dt7tau*(cij - ij) + sigma2*rng.normal(size=2)
            ij[:] = (ij + ij0).clip((0, 0), (ni, nj))

            # select image
            k = int((t-dt) / presentation_time + 1e-7)
            image = images[k % n]

            # interpolate jittered sub-image
            i, j = ij
            image = scipy.ndimage.interpolation.shift(
                image, (0, ni-i, nj-j))[:, -nyi:, -nyj:]

            return image.ravel()

        return step_presentjitteredimages



import numpy as np
import pytest

import nengo
from nengo.utils.stdlib import Timer

from nengo_extras.convnet import Conv2d, Pool2d


@pytest.mark.parametrize('local', [False, True])
def test_conv2d(local, Simulator, rng):
    f = 4
    c = 2
    ni, nj = 30, 32
    si, sj = 5, 3

    # f = 64
    # c = 64
    # ni, nj = 32, 32
    # si, sj = 11, 11

    si2 = int((si - 1) / 2.)
    sj2 = int((sj - 1) / 2.)

    fshape = (f, ni, nj, c, si, sj) if local else (f, c, si, sj)
    filters = rng.uniform(-1, 1, size=fshape)
    biases = rng.uniform(-1, 1, size=f)
    image = rng.uniform(-1, 1, size=(c, ni, nj))

    model = nengo.Network()
    with model:
        u = nengo.Node(image.ravel())
        v = nengo.Node(Conv2d(
            (c, ni, nj), filters, biases, padding=(si2, sj2)))
        nengo.Connection(u, v, synapse=None)
        vp = nengo.Probe(v)

    with Simulator(model) as sim:
        with Timer() as timer:
            sim.step()

    print("Conv2d(local=%s): %0.3e" % (local, timer.duration))

    # --- check result
    result = np.zeros((f, ni, nj))
    for i in range(ni):
        for j in range(nj):
            i0, i1 = i - si2, i + si2 + 1
            j0, j1 = j - sj2, j + sj2 + 1
            sli = slice(max(-i0, 0), min(ni + si - i1, si))
            slj = slice(max(-j0, 0), min(nj + sj - j1, sj))
            w = (filters[:, i, j, :, sli, slj] if local else
                 filters[:, :, sli, slj])
            xij = image[:, max(i0, 0):min(i1, ni), max(j0, 0):min(j1, nj)]
            result[:, i, j] += np.dot(xij.ravel(), w.reshape(f, -1).T)

    result += biases.reshape(-1, 1, 1)

    y = sim.data[vp][-1].reshape((f, ni, nj))
    assert np.allclose(result, y, rtol=1e-3, atol=1e-6)


@pytest.mark.parametrize('s, st', [(2, 2), (3, 1), (3, 3)])
def test_pool2d(s, st, Simulator, rng):
    nc = 3
    nxi, nxj = 30, 32
    image = rng.normal(size=(nc, nxi, nxj))

    # compute correct result
    nyi = 1 + int(np.ceil(float(nxi - s) / st))
    nyj = 1 + int(np.ceil(float(nxj - s) / st))
    nxi2, nxj2 = nyi * st, nyj * st
    result = np.zeros((nc, nyi, nyj))
    count = np.zeros((nyi, nyj))
    for i in range(s):
        for j in range(s):
            xij = image[:, i:min(nxi2+i, nxi):st, j:min(nxj2+j, nxj):st]
            ni, nj = xij.shape[-2:]
            result[:, :ni, :nj] += xij
            count[:ni, :nj] += 1

    result /= count

    # run simulation
    model = nengo.Network()
    with model:
        u = nengo.Node(image.ravel())
        v = nengo.Node(Pool2d(image.shape, s, strides=st))
        nengo.Connection(u, v, synapse=None)
        vp = nengo.Probe(v)

    with Simulator(model) as sim:
        sim.step()
    y = sim.data[vp][-1].reshape(result.shape)
    assert np.allclose(result, y, rtol=1e-3, atol=1e-6)



from nengo.dists import Gaussian, Uniform
import numpy as np

from nengo_extras.dists import (
    Concatenate,
    gaussian_icdf,
    loggaussian_icdf,
    Mixture,
    MultivariateCopula,
    MultivariateGaussian,
    Tile,
    uniform_icdf
)


def test_concatenate(plt, rng):
    n = 10000

    dist = Concatenate([Uniform(-1, 1),
                        Uniform(0, 1),
                        MultivariateGaussian([0, 2], [2, 1]),
                        Gaussian(3, 0.5)])
    pts = dist.sample(n, rng=rng)
    assert pts.shape == (n, 5)
    n, d = pts.shape

    for i in range(d):
        plt.subplot(d, 1, i+1)
        plt.hist(pts[:, i], bins=np.linspace(-4, 4, 101))


def test_icdfs(plt, rng):
    p = rng.rand(100000)

    rows = 3
    plt.subplot(rows, 1, 1)
    plt.hist(gaussian_icdf(1, 0.5)(p), bins=51)

    plt.subplot(rows, 1, 2)
    h, b = np.histogram(loggaussian_icdf(-2, 0.5, base=10)(p),
                        bins=np.logspace(-4, 0))
    plt.semilogx(0.5*(b[:-1] + b[1:]), h)

    plt.subplot(rows, 1, 3)
    plt.hist(uniform_icdf(-3, 1)(p), bins=51)


def test_multivariate_copula_simple(plt, rng):
    n = 100000

    c = 0.7
    dist = MultivariateCopula([gaussian_icdf(-1, 1),
                               uniform_icdf(-1, 1)],
                              rho=[[1., c], [c, 1.]])
    pts = dist.sample(n, rng=rng)
    assert pts.shape == (n, 2)

    plt.hist2d(pts[:, 0], pts[:, 1], bins=31)


def test_multivariate_gaussian(plt, rng):
    n = 200000

    dist = MultivariateGaussian((-1, 1), (0.25, 4))
    pts = dist.sample(n, rng=rng)

    plt.hist2d(pts[:, 0], pts[:, 1], bins=np.arange(-5, 5, 0.2))


def test_mixture_1d(plt, rng):
    n = 100000
    bins = np.arange(-5, 5, 0.1)

    dists = [Uniform(-0.5, 0.5), Gaussian(-1, 1), Gaussian(2, 1)]
    mdists = [Mixture(dists),
              Mixture(dists, p=[0.7, 0.15, 0.15])]

    for k, dist in enumerate(mdists):
        plt.subplot(len(mdists), 1, k+1)
        pts = dist.sample(n, rng=rng)
        plt.hist(pts, bins=bins)


def test_mixture_2d(plt, rng):
    n = 200000

    dists = [MultivariateGaussian((-2, -1), (0.25, 4)),
             MultivariateGaussian((1, 1), (1, 1))]
    mdists = [Mixture(dists),
              Mixture(dists, p=[0.7, 0.3])]

    for k, dist in enumerate(mdists):
        plt.subplot(len(mdists), 1, k+1)
        pts = dist.sample(n, d=2, rng=rng)
        plt.hist2d(pts[:, 0], pts[:, 1], bins=np.arange(-5, 5, 0.2))


def test_tile(rng):
    a = rng.uniform(-1, 1, size=(10, 3))

    dist = Tile(a)

    b = dist.sample(25, 4)
    assert np.array_equal(b, np.tile(a, (3, 2))[:25, :4])



from __future__ import absolute_import

import numpy as np

from nengo.dists import Distribution
from nengo.params import NdarrayParam, NumberParam, TupleParam


def gaussian_icdf(mean, std):
    import scipy.stats as sps

    def icdf(p):
        return sps.norm.ppf(p, scale=std, loc=mean)

    return icdf


def loggaussian_icdf(log_mean, log_std, base=np.e):
    import scipy.stats as sps

    mean = base**log_mean
    log_std2 = log_std * np.log(base)

    def icdf(p):
        return sps.lognorm.ppf(p, log_std2, scale=mean)

    return icdf


def uniform_icdf(low, high):
    def icdf(p):
        return p * (high - low) + low

    return icdf


class Concatenate(Distribution):
    """Concatenate distributions to form an independent multivariate"""

    distributions = TupleParam('distributions', readonly=True)
    d = NumberParam('d', low=1, readonly=True)

    def __init__(self, distributions):
        super(Concatenate, self).__init__()
        self.distributions = distributions

        # --- determine dimensionality
        rng = np.random.RandomState(0)
        s = np.column_stack([d.sample(1, rng=rng) for d in self.distributions])
        self.d = s.shape[1]

    def sample(self, n, d=None, rng=np.random):
        assert d is None or d == self.d
        return np.column_stack(
            [dist.sample(n, rng=rng) for dist in self.distributions])


class MultivariateCopula(Distribution):
    """Generalized multivariate distribution.

    Uses the copula method to sample from a general multivariate distribution,
    given marginal distributions and copula covariances [1]_.

    Parameters
    ----------
    marginal_icdfs : iterable
        List of functions, each one being the inverse CDF of the marginal
        distribution across that dimension.
    rho : array_like (optional)
        Array of copula covariances [1]_ between parameters. Defaults to
        the identity matrix (independent parameters).

    See also
    --------
    gaussian_icdf, loggaussian_icdf, uniform_icdf

    References
    ----------
    .. [1] Copula (probability theory). Wikipedia.
       https://en.wikipedia.org/wiki/Copula_(probability_theory%29
    """

    marginal_icdfs = TupleParam('marginal_icdfs', readonly=True)
    rho = NdarrayParam('rho', shape=('*', '*'), optional=True, readonly=True)

    def __init__(self, marginal_icdfs, rho=None):
        import scipy.stats  # we need this for sampling
        assert scipy.stats

        super(MultivariateCopula, self).__init__()
        self.marginal_icdfs = marginal_icdfs
        self.rho = rho

        d = len(self.marginal_icdfs)
        if not all(callable(f) for f in self.marginal_icdfs):
            raise ValueError("`marginal_icdfs` must be a list of callables")
        if self.rho is not None:
            if self.rho.shape != (d, d):
                raise ValueError("`rho` must be a %d x %d array" % (d, d))
            if not np.array_equal(self.rho, self.rho.T):
                raise ValueError(
                    "`rho` must be a symmetrical positive-definite array")

    def sample(self, n, d=None, rng=np.random):
        import scipy.stats as sps

        assert d is None or d == len(self.marginal_icdfs)
        d = len(self.marginal_icdfs)

        # normalize rho
        rho = np.eye(d) if self.rho is None else self.rho
        stds = np.sqrt(np.diag(rho))
        rho = rho / np.outer(stds, stds)

        # sample from copula
        x = sps.norm.cdf(sps.multivariate_normal.rvs(cov=rho, size=n))

        # apply marginal inverse CDFs
        for i in range(d):
            x[:, i] = self.marginal_icdfs[i](x[:, i])

        return x


class MultivariateGaussian(Distribution):
    mean = NdarrayParam('mean', shape='d')
    cov = NdarrayParam('cov', shape=('d', 'd'))

    def __init__(self, mean, cov):
        super(MultivariateGaussian, self).__init__()

        self.d = len(mean)
        self.mean = mean
        cov = np.asarray(cov)
        self.cov = (cov*np.eye(self.d) if cov.size == 1 else
                    np.diag(cov) if cov.ndim == 1 else cov)

    def sample(self, n, d=None, rng=np.random):
        assert d is None or d == self.d
        return rng.multivariate_normal(self.mean, self.cov, size=n)


class Mixture(Distribution):
    distributions = TupleParam('distributions')
    p = NdarrayParam('p', shape='*', optional=True)

    def __init__(self, distributions, p=None):
        super(Mixture, self).__init__()

        self.distributions = distributions
        if not all(isinstance(d, Distribution) for d in self.distributions):
            raise ValueError(
                "All elements in `distributions` must be Distributions")

        if p is not None:
            p = np.array(p)
            if p.ndim != 1 or p.size != len(self.distributions):
                raise ValueError(
                    "`p` must be a vector with one element per distribution")
            if (p < 0).any():
                raise ValueError("`p` must be all non-negative")
            p /= p.sum()
        self.p = p

    def sample(self, n, d=None, rng=np.random):
        dd = 1 if d is None else d
        samples = np.zeros((n, dd))

        ndims = len(self.distributions)
        i = (rng.randint(ndims, size=n) if self.p is None else
             rng.choice(ndims, p=self.p, size=n))
        c = np.bincount(i, minlength=ndims)

        for k in c.nonzero()[0]:
            samples[i == k] = self.distributions[k].sample(c[k], d=dd, rng=rng)

        return samples[:, 0] if d is None else samples


class Tile(Distribution):
    """Choose values in order from an array

    This distribution is not random, but rather tiles an array to be a
    particular size. This is useful for example if you want to pass an array
    for a neuron parameter, but are not sure how many neurons there will be.

    Parameters
    ----------
    values : array_like
        The values to tile.
    """

    values = NdarrayParam('values', shape=('*', '*'))

    def __init__(self, values):
        super(Tile, self).__init__()

        values = np.asarray(values)
        self.values = values.reshape(-1, 1) if values.ndim < 2 else values

    def __repr__(self):
        return "Tile(values=%s)" % (self.values)

    def sample(self, n, d=None, rng=np.random):
        out1 = d is None
        d = 1 if d is None else d
        nv, dv = self.values.shape

        if n > nv or d > dv:
            values = np.tile(self.values, (int(np.ceil(float(n) / nv)),
                                           int(np.ceil(float(d) / dv))))
        else:
            values = self.values

        values = values[:n, :d]
        return values[:, 0] if out1 else values



"""Creation of informative spike raster plots.

This module provides the `plot_spikes` function to easily create a spike raster
plot. Often such a plot is more informative with some preprocessing on the
spike trains that selects those with a high variance (i.e., changes in firing
rate) and orders them according to similarity. Such a preprocessing is provided
by the `preprocess_spikes` function. Thus, to create a quick spike raster plot
with such preprocessing::

    plot_spikes(*preprocess_spikes(t, spikes))

Functions for the individual preprocessing steps (and some alternatives) can
also be found in this model to allow to fine-tune the preprocessing according
to the data.
"""

from __future__ import absolute_import

import matplotlib
import matplotlib.pyplot as plt
import numpy as np


cm_gray_r_a = matplotlib.colors.LinearSegmentedColormap.from_list(
    'gray_r_a', [(0., 0., 0., 0.), (0., 0., 0., 1.)])


def plot_spikes(t, spikes, contrast_scale=1.0, ax=None, **kwargs):
    """Plots a spike raster.

    Will use an alpha channel by default which allows to plot colored regions
    below the spike raster to add highlights. You can set the *cmap* keyword
    argument to a different color map like *matplotlib.cm.gray_r* to get an
    opaque spike raster.

    Utilizes Matplotlib's *imshow*.

    Parameters
    ----------
    t : (n,) array
        Time indices of *spike* matrix. The indices are assumed to be
        equidistant.
    spikes : (n, m) array
        Spike data for *m* neurons at *n* time points.
    contrast_scale : float, optional
        Scales the contrst of the spike raster. This value multiplied with the
        maximum value in *spikes* will determine the minimum *spike* value to
        appear black (or the corresponding color of the chosen colormap).
    ax : matplotlib.axes.Axes, optional
        Axes to plot onto. Uses the current axes by default.
    kwargs : dict
        Additional keyword arguments will be passed on to *imshow*.

    Returns
    -------
    matplotlib.image.AxesImage
        The spikeraster.
    """

    t = np.asarray(t)
    spikes = np.asarray(spikes)
    if ax is None:
        ax = plt.gca()

    kwargs.setdefault('aspect', 'auto')
    kwargs.setdefault('cmap', cm_gray_r_a)
    kwargs.setdefault('interpolation', 'nearest')
    kwargs.setdefault('extent', (t[0], t[-1], 0., spikes.shape[1]))

    spikeraster = ax.imshow(spikes.T, **kwargs)
    spikeraster.set_clim(0., np.max(spikes) * contrast_scale)
    return spikeraster


def preprocess_spikes(
        t, spikes, num=50, sample_size=200, sample_filter_width=0.02,
        cluster_filter_width=0.002):
    """Applies a default preprocessing to spike data for plotting.

    This will first sample by variance, then cluster the spike trains, and
    finally merge them. See `sample_by_variance`, `cluster`, and `merge` for
    details.

    Parameters
    ----------
    t : (n,) array
        Time indices of *spike* matrix. The indices are assumed to be
        equidistant.
    spikes : (n, m) array
        Spike data for *m* neurons at *n* time points.
    num : int, optional
        Number of spike trains to return after merging.
    sample_size : int, optional
        Number of spike trains to sample by variance.
    sample_filter_width : float, optional
        Gaussian filter width in seconds for sampling by variance.
    cluster_filter_width : float, optional
        Gaussian filter width in seconds for clustering.

    Returns
    -------
    tuple (t, selected_spikes)
        Returns the time indices *t* and the preprocessed spike trains
        *spikes*.
    """
    return merge(
        *cluster(
            *sample_by_variance(
                t, spikes, num=sample_size, filter_width=sample_filter_width),
            filter_width=cluster_filter_width),
        num=num)


def cluster(t, spikes, filter_width):
    """Change order of spike trains to have similar ones close together.

    Requires SciPy.

    Parameters
    ----------
    t : (n,) array
        Time indices of *spike* matrix. The indices are assumed to be
        equidistant.
    spikes : (n, m) array
        Spike data for *m* neurons at *n* time points.
    filter_width : float
        Gaussian filter width in seconds, controls the time scale the
        clustering is sensitive to.

    Returns
    -------
    tuple (t, selected_spikes)
        Returns the time indices *t* and the selected spike trains *spikes*.
    """

    from scipy.cluster.hierarchy import linkage, to_tree
    from scipy.ndimage import gaussian_filter1d
    dt = (t[-1] - t[0]) / (len(t) - 1)
    filtered = gaussian_filter1d(np.asfarray(spikes),
                                 filter_width / dt, axis=0)
    order = to_tree(linkage(filtered.T)).pre_order()
    return t, spikes[:, order]


def merge(t, spikes, num):
    spikes = np.asarray(spikes)

    if spikes.shape[1] <= num:
        return t, spikes

    blocksize = int(np.ceil(spikes.shape[1] / num))
    merged = np.array([
        np.mean(spikes[:, (i * blocksize):((i + 1) * blocksize)], axis=1)
        for i in range(num)]).T
    return t, merged


def sample_by_variance(t, spikes, num, filter_width):
    """Samples the spike trains with the highest variance.

    Requires SciPy.

    Parameters
    ----------
    t : (n,) array
        Time indices of *spike* matrix. The indices are assumed to be
        equidistant.
    spikes : (n, m) array
        Spike data for *m* neurons at *n* time points.
    num : int
        Number of spike trains to return.
    filter_width : float
        Gaussian filter width in seconds, controls the time scale the variance
        calculation is sensitive to.

    Returns
    -------
    tuple (t, selected_spikes)
        Returns the time indices *t* and the selected spike trains *spikes*.
    """

    from scipy.ndimage import gaussian_filter1d
    dt = (t[-1] - t[0]) / (len(t) - 1)
    filtered = gaussian_filter1d(np.asfarray(spikes),
                                 filter_width / dt, axis=0)
    selected = np.argsort(np.var(filtered, axis=0))[-1:(-num - 1):-1]
    return t, spikes[:, selected]


def sample_by_activity(t, spikes, num, blocksize=None):
    """Samples the spike trains with the highest spiking activity.

    Parameters
    ----------
    t : (n,) array
        Time indices of *spike* matrix. The indices are assumed to be
        equidistant.
    spikes : (n, m) array
        Spike data for *m* neurons at *n* time points.
    num : int
        Number of spike trains to return.
    blocksize : int, optional
        If not *None*, the spike trains will be divided into blocks of this
        size and the highest activity spike trains are obtained for each block
        individually.

    Returns
    -------
    tuple (t, selected_spikes)
        Returns the time indices *t* and the selected spike trains *spikes*.
    """
    spikes = np.asarray(spikes)

    if spikes.shape[1] <= num:
        return t, spikes

    if blocksize is None:
        blocksize = spikes.shape[1]

    selected = np.empty((len(t), num))
    n_blocks = int(np.ceil(float(spikes.shape[1]) / blocksize))
    n_sel = int(np.ceil(float(num) / n_blocks))
    for i in range(n_blocks):
        block = spikes[:, (i * blocksize):((i + 1) * blocksize)]
        activity = np.sum(block, axis=0)
        selected[:, (i * n_sel):((i + 1) * n_sel)] = (
            block[:, np.argsort(activity)[-1:(-n_sel - 1):-1]])

    return t, selected



import numpy as np

import nengo
from nengo.dists import Choice
from nengo.networks.ensemblearray import EnsembleArray


def Product(n_neurons, dimensions, input_magnitude=1, net=None):
    """Computes the element-wise product of two equally sized vectors."""
    if net is None:
        net = nengo.Network(label="Product")

    config = nengo.Config(nengo.Ensemble)
    config[nengo.Ensemble].encoders = Choice(
        [[1, 1], [1, -1], [-1, 1], [-1, -1]])

    with net, config:
        net.A = nengo.Node(size_in=dimensions, label="A")
        net.B = nengo.Node(size_in=dimensions, label="B")
        net.product = EnsembleArray(n_neurons, n_ensembles=dimensions,
                                    ens_dimensions=2,
                                    radius=input_magnitude * np.sqrt(2))
        nengo.Connection(net.A, net.product.input[::2], synapse=None)
        nengo.Connection(net.B, net.product.input[1::2], synapse=None)
        net.output = net.product.add_output('product', lambda x: x[0] * x[1])
    return net



import nengo
from nengo.utils.compat import is_iterable


def probe_all(  # noqa: C901
        net, recursive=False, probe_options=None, **probe_args):
    """Probes all objects in a network.

    Parameters
    ----------
    net : nengo.Network
    recursive : bool, optional (Default: False)
        Probe subnetworks recursively.
    probe_options: dict, optional (Default: None)
        A dict of the form {nengo_object_class: [attributes_to_probe]}.
        If None, every probeable attribute of every object will be probed.

    Returns
    -------
    A dictionary that maps objects and their attributes to their probes.

    Examples
    --------

    Probe the decoded output and spikes in all ensembles in a network and
    its subnetworks::

        with nengo.Network() as model:
            ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
            node1 = nengo.Node(output=[0])
            conn = nengo.Connection(node1, ens1)
            subnet = nengo.Network(label='subnet')
            with subnet:
                ens2 = nengo.Ensemble(n_neurons=1, dimensions=1)
                node2 = nengo.Node(output=[0])
        probe_options = {nengo.Ensemble: ['decoded_output', 'spikes']}
        probes = probe_all(model, recursive=True, probe_options=probe_options)

    """

    probes = {}

    def all_probes(obj):
        if probe_options is not None and type(obj) in probe_options:
            attrs = probe_options[type(obj)]
        else:
            attrs = obj.probeable

        return {attr: nengo.Probe(obj, attr, **probe_args) for attr in attrs}

    ensembles = net.all_ensembles if recursive else net.ensembles
    nodes = net.all_nodes if recursive else net.nodes
    connections = net.all_connections if recursive else net.connections

    with net:
        if probe_options is None or nengo.Ensemble in probe_options:
            for ens in ensembles:
                probes[ens] = all_probes(ens)

        if probe_options is None or nengo.ensemble.Neurons in probe_options:
            for ens in ensembles:
                probes[ens.neurons] = all_probes(ens.neurons)

        if probe_options is None or nengo.Node in probe_options:
            for node in nodes:
                probes[node] = all_probes(node)

        if probe_options is None or nengo.Connection in probe_options:
            for conn in connections:
                probes[conn] = all_probes(conn)

        LearningRule = nengo.connection.LearningRule
        if probe_options is None or LearningRule in probe_options:
            for conn in connections:
                lr = conn.learning_rule
                if lr is None:
                    continue
                if not is_iterable(lr):
                    lr = [lr]
                for rule in lr:
                    probes[rule] = all_probes(rule)

    return probes



"""
Use Nengo's RC files to manage our RC parameters.
"""
import types

from nengo import rc


def rc_setdefault(self, section, option, value=None):
    if not self.has_section(section):
        self.add_section(section)
    if not self.has_option(section, option):
        self.set(section, option, value=value)


rc.setdefault = types.MethodType(rc_setdefault, rc)

# --- set our defaults
rc.setdefault('nengo_extras', 'data_dir', '.')



"""
Tools for running deep networks in Nengo

Notes:
- Layers with weights make copies of said weights, both so that they are
  independent (if the model is later updated), and so they are contiguous.
  (Contiguity is required in Nengo when arrays are hashed during build.)
"""

from __future__ import absolute_import

import numpy as np

import nengo
from nengo.params import Default
from nengo.utils.network import with_self

from .convnet import Conv2d, Pool2d, softmax
from .neurons import SoftLIFRate


class Network(nengo.Network):
    @with_self
    def add_data_layer(self, d, name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = DataLayer(d, **kwargs)
        return self.add_layer(layer, inputs=(), name=name)

    @with_self
    def add_neuron_layer(self, n, inputs=None, name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = NeuronLayer(n, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_softmax_layer(self, size, inputs=None, name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = SoftmaxLayer(size, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_dropout_layer(self, size, keep, inputs=None, name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = DropoutLayer(size, keep, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_full_layer(self, weights, biases, inputs=None, name=None,
                       **kwargs):
        kwargs.setdefault('label', name)
        layer = FullLayer(weights, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_local_layer(self, input_shape, filters, biases, inputs=None,
                        name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = LocalLayer(input_shape, filters, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_conv_layer(self, input_shape, filters, biases, inputs=None,
                       name=None, **kwargs):
        kwargs.setdefault('label', name)
        layer = ConvLayer(input_shape, filters, biases, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    @with_self
    def add_pool_layer(self, input_shape, pool_size, inputs=None, name=None,
                       **kwargs):
        kwargs.setdefault('label', name)
        layer = PoolLayer(input_shape, pool_size, **kwargs)
        return self.add_layer(layer, inputs=inputs, name=name)

    def compute(self, inputs, output):
        raise NotImplementedError()


class SequentialNetwork(Network):
    def __init__(self, **kwargs):
        super(SequentialNetwork, self).__init__(**kwargs)

        self.layers = []
        self.layers_by_name = {}

    @property
    def input(self):
        return (None if len(self.layers) == 0 else self.layers[0].input)

    @property
    def output(self):
        return (None if len(self.layers) == 0 else self.layers[-1].output)

    @with_self
    def add_layer(self, layer, inputs=None, name=None):
        assert isinstance(layer, Layer)
        assert layer not in self.layers

        if inputs is None:
            inputs = [self.layers[-1]] if self.output is not None else []

        assert len(inputs) == 0 if self.output is None else (
            len(inputs) == 1 and inputs[0] is self.layers[-1])

        for i in inputs:
            nengo.Connection(
                i.output, layer.input, synapse=None, **layer.pre_args)
        self.layers.append(layer)

        if name is not None:
            assert name not in self.layers_by_name
            self.layers_by_name[name] = layer

        return layer

    def layers_to(self, end):
        if end is None:
            return self.layers

        assert end in self.layers
        return self.layers[:self.layers.index(end)+1]

    def compute(self, x, output_layer=None):
        y = x
        for layer in self.layers_to(output_layer):
            y = layer.compute(y)

        return y

    def theano(self, sx, output_layer=None):
        # ensure we have Theano
        import theano

        sy = sx
        for layer in self.layers_to(output_layer):
            sy = layer.theano(sy)

        return sy

    def theano_compute(self, x, output_layer=None, batch_size=256):
        import theano
        import theano.tensor as tt
        sx = tt.matrix(name='x')
        sy = self.theano(sx, output_layer=output_layer)
        f = theano.function([sx], sy, allow_input_downcast=True)

        x = x.reshape((x.shape[0], -1))
        y0 = f(x[:batch_size])
        if len(x) == len(y0):
            return y0
        else:
            assert len(x) > len(y0)
            y = np.zeros((len(x), y0.shape[1]), dtype=y0.dtype)
            y[:batch_size] = y0
            for i in range(batch_size, len(x), batch_size):
                y[i:i+batch_size] = f(x[i:i+batch_size])
            return y


# class TreeNetwork(Network):
#     def __init__(self, **kwargs):
#         super(TreeNetwork, self).__init__(**kwargs)

#         self.inputs = {}  # name -> input object
#         self.outputs = {}  # name -> output object
#         self.layer_inputs = {}  # mapping layer to its input layers
#         self.layers_by_name = {}

#     @property
#     def layers(self):
#         return list(self.layer_inputs)

#     @with_self
#     def add_layer(self, inputs, layer, name=None):
#         assert isinstance(layer, Layer)
#         assert layer not in self.layer_inputs
#         for i in inputs:
#             assert i in self.layer_inputs
#             nengo.Connection(
#                 i.output, layer.input, synapse=None, **layer.pre_args)

#         self.layer_inputs[layer] = tuple(inputs)
#         if name is not None:
#             assert name not in self.layers_by_name
#             self.layers_by_name[name] = layer

#         return layer

#     @with_self
#     def add_named_input(self, name, d, **kwargs):
#         kwargs.setdefault('label', name)
#         layer = DataLayer(d, **kwargs)
#         self.inputs[name] = layer.input
#         return self.add_layer(layer, inputs=(), name=name)

#     def add_named_output(self, name, obj):
#         output = obj.output
#         self.outputs[name] = output
#         return output

#     def compute(self, inputs, output):
#         raise NotImplementedError(
#             "Layers currently only handle one input for compute")

#         assert isinstance(inputs, dict)

#         def as_layer(x):
#             if isinstance(x, str):
#                 return self.layers_by_name[x]
#             else:
#                 assert x in self.layer_inputs
#                 return x

#         if any(isinstance(i, str) for i in inputs):
#             inputs = dict((as_layer(k), v) for k, v in inputs.items())
#         else:
#             assert all(i in self.layer_inputs for i in inputs)
#         output = as_layer(output)

#         layer_inputs = [inputs[i] if i in inputs else self.compute(inputs, i)
#                         for i in self.layer_inputs[output]]
#         assert len(layer_inputs) == 1
#         return output.compute(layer_inputs[0])


class Layer(nengo.Network):
    def __init__(self, **kwargs):
        super(Layer, self).__init__(**kwargs)
        self.pre_args = {}  # kwargs for incoming connections
        # self.post_args = {}  # kwargs for outgoing connections

    @property
    def input(self):
        raise NotImplementedError()

    @property
    def output(self):
        raise NotImplementedError()

    @property
    def size_in(self):
        return self.input.size_in

    @property
    def size_out(self):
        return self.output.size_out

    @property
    def shape_in(self):
        return (self.size_in,)

    @property
    def shape_out(self):
        return (self.size_out,)

    def compute(self, x):
        raise NotImplementedError()

    def theano(self, sx):
        raise NotImplementedError()

    def theano_compute(self, x, **kwargs):
        import theano
        import theano.tensor as tt
        sx = tt.matrix()
        f = theano.function([sx], self.theano(sx), **kwargs)
        y = f(x)
        return y

    def _compute_input(self, x):
        x = np.asarray(x)
        if x.ndim == 0:
            raise ValueError("'x' must be at least 1D")
        elif x.ndim == 1:
            assert x.shape[0] == self.size_in
            return x.reshape(1, -1)

        assert x.ndim == 2 and x.shape[1] == self.size_in
        return x


class NodeLayer(Layer):
    def __init__(self, output=None, size_in=None, **kwargs):
        super(NodeLayer, self).__init__(**kwargs)

        with self:
            self.node = nengo.Node(output=output, size_in=size_in)

        if self.label is not None:
            self.node.label = '%s_node' % self.label

    @property
    def input(self):
        return self.node

    @property
    def output(self):
        return self.node


class NeuronLayer(Layer):
    def __init__(self, n, neuron_type=Default, synapse=Default,
                 gain=1., bias=0., amplitude=1., **kwargs):
        super(NeuronLayer, self).__init__(**kwargs)

        with self:
            self.ensemble = nengo.Ensemble(n, 1, neuron_type=neuron_type)
            self.ensemble.gain = _vectorize(gain, n, 'gain')
            self.ensemble.bias = _vectorize(bias, n, 'bias')

            self._output = nengo.Node(size_in=n)
            self.connection = nengo.Connection(
                self.ensemble.neurons, self.output, transform=amplitude,
                synapse=synapse)

        if self.label is not None:
            self.ensemble.label = '%s_ensemble' % self.label
            self.output.label = '%s_output' % self.label

    @property
    def input(self):
        return self.ensemble.neurons

    @property
    def output(self):
        return self._output

    @property
    def neurons(self):
        return self.ensemble.neurons

    @property
    def amplitude(self):
        return self.connection.transform

    @property
    def bias(self):
        return self.ensemble.bias

    @property
    def gain(self):
        return self.ensemble.gain

    @property
    def neuron_type(self):
        return self.ensemble.neuron_type

    @property
    def synapse(self):
        return self.connection.synapse

    def compute(self, x):
        x = self._compute_input(x)
        return self.amplitude * self.neuron_type.rates(x, self.gain, self.bias)

    def theano(self, sx):
        import theano
        import theano.tensor as tt
        floatX = theano.config.floatX
        gain = tt.cast(self.gain, floatX)
        bias = tt.cast(self.bias, floatX)
        amp = tt.cast(self.amplitude, floatX)
        return amp * neuron_theano(self.neuron_type, gain*sx + bias)


class DataLayer(NodeLayer):
    def __init__(self, size, **kwargs):
        super(DataLayer, self).__init__(size_in=size, **kwargs)

    def compute(self, x):
        return x.reshape((x.shape[0], self.size_out))

    def theano(self, sx):
        return sx.reshape((sx.shape[0], self.size_out))


class SoftmaxLayer(NodeLayer):
    def __init__(self, size, **kwargs):
        super(SoftmaxLayer, self).__init__(
            output=lambda t, x: softmax(x), size_in=size, **kwargs)

    def compute(self, x):
        x = self._compute_input(x)
        return softmax(x, axis=-1)

    def theano(self, sx):
        import theano.tensor as tt
        assert sx.ndim == 2
        return tt.nnet.softmax(sx)


class DropoutLayer(NodeLayer):
    def __init__(self, size, keep, **kwargs):
        super(DropoutLayer, self).__init__(size_in=size, **kwargs)
        self.pre_args['transform'] = keep

    @property
    def keep(self):
        return self.pre_args['transform']

    def compute(self, x):
        x = self._compute_input(x)
        return self.keep * x

    def theano(self, sx):
        return self.keep * sx


class FullLayer(NodeLayer):
    def __init__(self, weights, biases, **kwargs):
        assert weights.ndim == 2
        assert biases.size == weights.shape[0]
        super(FullLayer, self).__init__(size_in=weights.shape[0], **kwargs)

        self.weights = np.array(weights)  # copy
        self.biases = np.array(biases)  # copy

        with self:
            self.bias = nengo.Node(output=biases)
            nengo.Connection(self.bias, self.node, synapse=None)

        self.pre_args['transform'] = weights

        if self.label is not None:
            self.bias.label = '%s_bias' % self.label

    def compute(self, x):
        x = self._compute_input(x)
        return np.dot(x, self.weights.T) + self.biases

    def theano(self, sx):
        import theano.tensor as tt
        assert sx.ndim == 2
        return tt.dot(sx, self.weights.T) + self.biases


class ProcessLayer(NodeLayer):
    def __init__(self, process, **kwargs):
        assert isinstance(process, nengo.Process)
        super(ProcessLayer, self).__init__(output=process, **kwargs)
        self._step = None

    @property
    def process(self):
        return self.node.output

    @property
    def shape_in(self):
        return self.process.shape_in

    @property
    def shape_out(self):
        return self.process.shape_out

    def compute(self, x):
        x = self._compute_input(x)
        n = x.shape[0]
        if self._step is None:
            self._step = self.process.make_step(
                self.size_in, self.size_out, dt=1., rng=None)
        return self._step(0, x).reshape(n, -1)


class LocalLayer(ProcessLayer):
    def __init__(self, input_shape, filters, biases,
                 strides=1, padding=0, **kwargs):
        assert filters.ndim == 6
        filters = np.array(filters)  # copy
        biases = np.array(biases)  # copy
        p = Conv2d(input_shape, filters, biases,
                   strides=strides, padding=padding)
        super(LocalLayer, self).__init__(p, **kwargs)

    def theano(self, x):
        import theano.tensor as tt

        filters = self.process.filters
        biases = self.process.biases
        nc, nxi, nxj = self.process.shape_in
        nf, nyi, nyj = self.process.shape_out
        si, sj = self.process.filters.shape[-2:]
        pi, pj = self.process.padding
        sti, stj = self.process.strides

        ys = []
        n = x.shape[0]
        x = x.reshape((n, nc, nxi, nxj))
        for i in range(nyi):
            for j in range(nyj):
                i0 = i*sti - pi
                j0 = j*stj - pj
                i1, j1 = i0 + si, j0 + sj
                sli = slice(max(-i0, 0), min(nxi + si - i1, si))
                slj = slice(max(-j0, 0), min(nxj + sj - j1, sj))
                w = filters[:, i, j, :, sli, slj].reshape(nf, -1)
                xij = x[:, :, max(i0, 0):min(i1, nxi), max(j0, 0):min(j1, nxj)]
                ys.append(tt.dot(xij.reshape((1, n, -1)), w.T))

        y = tt.concatenate(ys, axis=0).reshape((nyi, nyj, n, nf))
        y = tt.transpose(y, (2, 3, 0, 1))

        y += biases
        return y.reshape((n, nf*nyi*nyj))


class ConvLayer(ProcessLayer):
    def __init__(self, input_shape, filters, biases,
                 strides=1, padding=0, border='ceil', **kwargs):
        assert filters.ndim == 4
        filters = np.array(filters)  # copy
        biases = np.array(biases)  # copy
        p = Conv2d(input_shape, filters, biases,
                   strides=strides, padding=padding, border=border)
        super(ConvLayer, self).__init__(p, **kwargs)

    def theano(self, x):
        import theano.tensor as tt

        filters = self.process.filters
        biases = self.process.biases
        nc, nxi, nxj = self.process.shape_in
        nf, nyi, nyj = self.process.shape_out
        si, sj = self.process.filters.shape[-2:]
        pi, pj = self.process.padding
        sti, stj = self.process.strides

        n = x.shape[0]
        x = x.reshape((n, nc, nxi, nxj))

        nxi2 = nyi*sti + si - 2*pi - 1
        nxj2 = nyj*stj + sj - 2*pj - 1
        if self.process.border == 'ceil' and (nxi2 > nxi or nxj2 > nxj):
            xx = tt.zeros((n, nc, nxi2, nxj2), dtype=x.dtype)
            x = tt.set_subtensor(xx[:, :, :nxi, :nxj], x)
        else:
            assert nxi == nxi2 and nxj == nxj2

        y = tt.nnet.conv2d(x, filters,
                           input_shape=(None, nc, nxi2, nxj2),
                           filter_shape=(nf, nc, si, sj),
                           border_mode=(pi, pj),
                           subsample=(sti, stj),
                           filter_flip=False)

        y += biases
        return y.reshape((n, nf*nyi*nyj))


class PoolLayer(ProcessLayer):
    def __init__(self, input_shape, pool_size,
                 strides=None, kind='avg', mode='full', **kwargs):
        p = Pool2d(input_shape, pool_size, strides=strides,
                   kind=kind, mode=mode)
        super(PoolLayer, self).__init__(p, **kwargs)

    def theano(self, x):
        import theano.tensor as tt
        import theano.tensor.signal.pool

        pool_size = self.process.pool_size
        strides = self.process.strides
        nc, nxi, nxj = self.process.shape_in
        nc, nyi, nyj = self.process.shape_out
        mode = dict(max='max', avg='average_exc_pad')[self.process.kind]

        n = x.shape[0]
        x = x.reshape((n, nc, nxi, nxj))
        y = tt.signal.pool.pool_2d(
            x, pool_size, ignore_border=False, st=strides, mode=mode)

        return y.reshape((n, nc*nyi*nyj))


# class NEFLayer(nengo.Network):
#     def __init__(self, n, d_in, d_out, synapse=Default, eval_points=Default,
#                  function=Default, solver=Default,
#                  label=None, seed=None, add_to_container=None, **ens_kwargs):
#         super(nengo.Network, self).__init__(
#             label=label, seed=seed, add_to_container=add_to_container)

#         with self:
#             self.ensemble = nengo.Ensemble(n, d_in, **ens_kwargs)
#             self.output = nengo.Node(size_in=d_out)
#             self.connection = nengo.Connection(
#                 self.ensemble, self.output, synapse=synapse,
#                 eval_points=eval_points, function=function, solver=solver)

#         if label is not None:
#             self.ensemble.label = '%s_ensemble' % label
#             self.output.label = '%s_output' % label

#     @property
#     def input(self):
#         return self.ensemble.neurons

#     @property
#     def amplitude(self):
#         return self.connection.transform

#     @property
#     def neuron_type(self):
#         return self.ensemble.neuron_type

#     @property
#     def synapse(self):
#         return self.connection.synapse


# def _get_network_attr(obj, attr):
#     if isinstance(obj, nengo.Network) and hasattr(obj, attr):
#         return getattr(obj, attr)
#     else:
#         return obj


def _vectorize(value, n, name):
    value = np.asarray(value)
    if value.size == 1:
        return value * np.ones(n)
    elif value.size == n:
        return value
    else:
        raise ValueError("%r must be length 1 or %d (got %d)"
                         % (name, n, value.size))


def neuron_theano(neuron_type, x):
    import theano
    import theano.tensor as tt
    floatX = theano.config.floatX

    if isinstance(neuron_type, nengo.Direct):
        return x
    elif isinstance(neuron_type, nengo.neurons.RectifiedLinear):
        return tt.maximum(x, 0)
    elif isinstance(neuron_type, nengo.neurons.Sigmoid):
        return tt.nnet.sigmoid(x)
    elif isinstance(neuron_type, SoftLIFRate):  # before LIF since subclass
        # do not apply amplitude, since this is done elsewhere!
        tau_ref = tt.cast(neuron_type.tau_ref, floatX)
        tau_rc = tt.cast(neuron_type.tau_rc, floatX)
        sigma = tt.cast(neuron_type.sigma, floatX)
        j = tt.nnet.softplus((x - 1) / sigma) * sigma
        r = 1 / (tau_ref + tau_rc*tt.log1p(1/j))
        return tt.switch(j > 0, r, 0)
    elif isinstance(neuron_type, (nengo.LIF, nengo.LIFRate)):
        tau_ref = tt.cast(neuron_type.tau_ref, floatX)
        tau_rc = tt.cast(neuron_type.tau_rc, floatX)
        j = x - 1
        r = 1. / (tau_ref + tau_rc*tt.log1p(1./j))
        return tt.switch(j > 0, r, 0)
    else:
        raise NotImplementedError("Neuron type %r" % neuron_type)



from __future__ import absolute_import

import base64
from io import BytesIO


def preprocess_display(x, transpose=(1, 2, 0), scale=255., offset=0.):
    """Basic preprocessing that reshapes, transposes, and scales an image"""
    x = (x + offset) * scale
    if transpose is not None:
        x = x.transpose(transpose)  # color channel last
    if x.shape[-1] == 1:
        x = x[..., 0]
    return x.clip(0, 255).astype('uint8')


def image_function(image_shape, preprocess=preprocess_display,
                   **preprocess_args):
    """Make a function to turn an array into a PIL Image.

    Parameters
    ----------
    image_shape : array_like (3,)
        The shape of the image: (channels, height, width)
    preprocess : callable
        Callable that takes an image and preprocesses it to be displayed.
    preprocess_args : dict
        Optional dictionary of keyword arguments for ``preprocess``.

    Returns
    -------
    to_pil : callable (x)
        A function that takes a (flattened) image array, returning a PIL Image.

    Requirements
    ------------
    pillow (provides PIL, `pip install pillow`)
    """
    import PIL.Image
    assert len(image_shape) == 3

    def to_pil(x):
        x = x.reshape(image_shape)
        y = preprocess(x, **preprocess_args)
        image = PIL.Image.fromarray(y)
        return image

    return to_pil


def image_string_function(image_shape, format="PNG",
                          preprocess=preprocess_display, **preprocess_args):
    """Make a function to turn an array into an image string.

    Parameters
    ----------
    image_shape : array_like (3,)
        The shape of the image: (channels, height, width)
    format : string
        A format string for the ``PIL.Image.save`` function.
    preprocess : callable
        Callable that takes an image and preprocesses it to be displayed.
    preprocess_args : dict
        Optional dictionary of keyword arguments for ``preprocess``.

    Returns
    -------
    string_function : callable (x)
        A function that takes a (flattened) image array, and returns a
        base64 string representation of the image of the requested format.

    See also
    --------
    image_function
    """
    to_pil = image_function(
        image_shape, preprocess=preprocess, **preprocess_args)

    def string_function(x):
        image = to_pil(x)
        buffer = BytesIO()
        image.save(buffer, format=format)
        image_string = base64.b64encode(buffer.getvalue()).decode()
        return image_string

    return string_function


def image_html_function(image_shape, preprocess=preprocess_display,
                        **preprocess_args):
    """Make a function to turn an image into HTML to display as an SVG.

    Parameters
    ----------
    image_shape : array_like (3,)
        The shape of the image: (channels, height, width)
    preprocess : callable
        Callable that takes an image and preprocesses it to be displayed.
    preprocess_args : dict
        Optional dictionary of keyword arguments for ``preprocess``.

    Returns
    -------
    html_function : callable (x)
        A function that takes a (flattened) image array, and returns
        a string that defines an SVG object in HTML to display the image.

    See also
    --------
    image_function
    """
    string_function = image_string_function(
        image_shape, preprocess=preprocess, **preprocess_args)

    def html_function(x):
        image_string = string_function(x)
        return '''
            <svg width="100%%" height="100%%" viewbox="0 0 100 100">
            <image width="100%%" height="100%%"
                   xlink:href="data:image/png;base64,%s"
                   style="image-rendering: pixelated;">
            </svg>''' % (''.join(image_string))

    return html_function


def image_display_function(image_shape, preprocess=preprocess_display,
                           **preprocess_args):
    """Make a function to display images in Nengo GUI

    Examples
    --------
    >>> u = nengo.Node(nengo.processes.PresentInput(images, 0.1))
    >>> display_f = nengo_extras.gui.image_display_function(image_shape)
    >>> display_node = nengo.Node(display_f, size_in=u.size_out)
    >>> nengo.Connection(u, display_node, synapse=None)

    Requirements
    ------------
    pillow (provides PIL, `pip install pillow`)
    """
    html_function = image_html_function(
        image_shape, preprocess=preprocess, **preprocess_args)

    def display_func(t, x):
        display_func._nengo_html_ = html_function(x)

    return display_func



from __future__ import absolute_import
import os

import keras
import nengo
import numpy as np

import nengo_extras.deepnetworks


class SoftLIF(keras.layers.Layer):
    def __init__(self, sigma=1., amplitude=1., tau_rc=0.02, tau_ref=0.002,
                 noise_model='none', tau_s=0.005, **kwargs):
        self.supports_masking = True
        self.sigma = sigma
        self.amplitude = amplitude
        self.tau_rc = tau_rc
        self.tau_ref = tau_ref

        if noise_model not in ('none', 'alpharc'):
            raise ValueError("Unrecognized noise model")
        self.noise_model = noise_model

        self.tau_s = tau_s
        if abs(self.tau_rc - self.tau_s) < 1e-4:
            raise ValueError("tau_rc and tau_s must be different")

        super(SoftLIF, self).__init__(**kwargs)

    def call(self, x, mask=None):
        from keras import backend as K
        if K.backend() == 'tensorflow':
            import tensorflow as tf
            expm1 = tf.expm1
            log1p = tf.log1p
            where = tf.where
        else:
            import theano.tensor as tt
            expm1 = tt.expm1
            log1p = tt.log1p
            where = tt.switch

        # compute non-noisy output
        xs = x / self.sigma
        x_valid = xs > -20
        xs_safe = where(x_valid, xs, K.zeros_like(xs))
        j = K.softplus(xs_safe) * self.sigma
        p = self.tau_ref + self.tau_rc*where(
            x_valid, log1p(1/j), -xs - np.log(self.sigma))
        r = self.amplitude/p

        if self.noise_model == 'none':
            return r
        elif self.noise_model == 'alpharc':
            # compute noisy output for forward pass
            d = self.tau_rc - self.tau_s
            u01 = K.random_uniform(K.shape(p))
            t = u01 * p
            q_rc = K.exp(-t / self.tau_rc)
            q_s = K.exp(-t / self.tau_s)
            r_rc1 = -expm1(-p / self.tau_rc)  # 1 - exp(-p/tau_rc)
            r_s1 = -expm1(-p / self.tau_s)  # 1 - exp(-p/tau_s)

            pt = where(p < 100*self.tau_s, (p - t)*(1 - r_s1), K.zeros_like(p))
            qt = where(t < 100*self.tau_s, q_s*(t + pt), K.zeros_like(t))
            rt = qt / (self.tau_s * d * r_s1**2)
            rn = self.tau_rc*(q_rc/(d*d*r_rc1) - q_s/(d*d*r_s1)) - rt

            # r + stop_gradient(rn - r) = rn on forward pass, r on backwards
            return r + K.stop_gradient(self.amplitude*rn - r)

    def get_config(self):
        config = {'sigma': self.sigma, 'amplitude': self.amplitude,
                  'tau_rc': self.tau_rc, 'tau_ref': self.tau_ref}
        base_config = super(SoftLIF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


_custom_objects = {
    'SoftLIF': SoftLIF
}


def load_model_pair(filepath, custom_objects=None):
    if custom_objects is None:
        custom_objects = {}

    json_path = filepath + '.json'
    h5_path = filepath + '.h5'

    combined_customs = dict(_custom_objects)
    combined_customs.update(custom_objects)

    with open(json_path, 'r') as f:
        model = keras.models.model_from_json(
            f.read(), custom_objects=combined_customs)

    model.load_weights(h5_path)
    return model


def save_model_pair(model, filepath, overwrite=False):
    json_path = filepath + '.json'
    h5_path = filepath + '.h5'

    if not overwrite and os.path.exists(json_path):
        raise ValueError("Path already exists: %r" % filepath)

    json_string = model.to_json()
    with open(json_path, 'w') as f:
        f.write(json_string)

    model.save_weights(h5_path, overwrite=overwrite)


def kmodel_compute_shapes(kmodel, input_shape):
    assert isinstance(kmodel, keras.models.Sequential)

    shapes = [input_shape]
    for layer in kmodel.layers:
        s = layer.compute_output_shape(shapes[-1])
        shapes.append(s)

    return shapes


class SequentialNetwork(nengo_extras.deepnetworks.SequentialNetwork):

    def __init__(self, model, synapse=None, lif_type='lif', **kwargs):
        super(SequentialNetwork, self).__init__(**kwargs)

        assert isinstance(model, keras.models.Sequential)
        self.model = model
        self.synapse = synapse
        self.lif_type = lif_type

        # -- build model
        self.add_data_layer(np.prod(model.input_shape[1:]))
        for layer in model.layers:
            self._add_layer(layer)

    def _add_layer(self, layer):
        assert layer.input_mask is None
        assert layer.input_shape[0] is None

        layer_adder = {
            keras.layers.Activation: self._add_activation_layer,
            keras.layers.Dense: self._add_dense_layer,
            keras.layers.Dropout: self._add_dropout_layer,
            keras.layers.Flatten: self._add_flatten_layer,
            keras.layers.Convolution2D: self._add_conv2d_layer,
            keras.layers.AveragePooling2D: self._add_avgpool2d_layer,
            keras.layers.MaxPooling2D: self._add_maxpool2d_layer,
            keras.layers.noise.GaussianNoise: self._add_gaussian_noise_layer,
            SoftLIF: self._add_softlif_layer,
        }

        for cls in type(layer).__mro__:
            if cls in layer_adder:
                return layer_adder[cls](layer)

        raise NotImplementedError("Cannot build layer type %r" %
                                  type(layer).__name__)

    def _add_dense_layer(self, layer):
        weights, biases = layer.get_weights()
        return self.add_full_layer(weights.T, biases, name=layer.name)

    def _add_conv2d_layer(self, layer):
        import keras.backend as K
        shape_in = layer.input_shape[1:]
        filters, biases = layer.get_weights()
        strides = layer.strides

        assert layer.data_format == 'channels_first'
        nc, _, _ = shape_in
        filters = np.transpose(filters, (3, 2, 0, 1))
        if K.backend() == 'theano':
            filters = filters[:, :, ::-1, ::-1]  # theano has flipped filters
        filters = filters.copy()  # to make contiguous

        nf, nc2, si, sj = filters.shape
        assert nc == nc2, "Filter channels must match input channels"

        if layer.padding == 'valid':
            padding = (0, 0)
        elif layer.padding == 'same':
            padding = ((si - 1) / 2, (sj - 1) / 2)
        else:
            raise ValueError("Unrecognized padding %r" % layer.padding)

        conv = self.add_conv_layer(
            shape_in, filters, biases, strides=strides, padding=padding,
            border='floor', name=layer.name)
        assert conv.size_out == np.prod(layer.output_shape[1:])
        return conv

    def _add_pool2d_layer(self, layer, kind=None):
        shape_in = layer.input_shape[1:]
        pool_size = layer.pool_size
        strides = layer.strides
        return self.add_pool_layer(shape_in, pool_size, strides=strides,
                                   kind=kind, mode='valid', name=layer.name)

    def _add_avgpool2d_layer(self, layer):
        return self._add_pool2d_layer(layer, kind='avg')

    def _add_maxpool2d_layer(self, layer):
        return self._add_pool2d_layer(layer, kind='max')

    def _add_activation_layer(self, layer):
        if layer.activation is keras.activations.softmax:
            return self._add_softmax_layer(layer)

        # add normal activation layer
        activation_map = {
            keras.activations.relu: nengo.neurons.RectifiedLinear(),
            keras.activations.sigmoid: nengo.neurons.Sigmoid(),
            }
        neuron_type = activation_map.get(layer.activation, None)
        if neuron_type is None:
            raise ValueError("Unrecognized activation type %r"
                             % layer.activation)

        n = np.prod(layer.input_shape[1:])
        return self.add_neuron_layer(
            n, neuron_type=neuron_type, synapse=self.synapse,
            gain=1, bias=0, name=layer.name)

    def _add_softlif_layer(self, layer):
        from .neurons import SoftLIFRate

        taus = dict(tau_rc=layer.tau_rc, tau_ref=layer.tau_ref)
        lif_type = self.lif_type.lower()
        if lif_type == 'lif':
            neuron_type = nengo.LIF(**taus)
        elif lif_type == 'lifrate':
            neuron_type = nengo.LIFRate(**taus)
        elif lif_type == 'softlifrate':
            neuron_type = SoftLIFRate(sigma=layer.sigma, **taus)
        else:
            raise KeyError("Unrecognized LIF type %r" % self.lif_type)

        n = np.prod(layer.input_shape[1:])
        return self.add_neuron_layer(
            n, neuron_type=neuron_type, synapse=self.synapse,
            gain=1, bias=1, amplitude=layer.amplitude, name=layer.name)

    def _add_softmax_layer(self, layer):
        return None  # non-neural, we can do without it
        # return self.add_softmax_layer(
        #     np.prod(layer.input_shape[1:]), name=layer.name)

    def _add_dropout_layer(self, layer):
        return None  # keras scales by dropout rate, so we don't have to

    def _add_flatten_layer(self, layer):
        return None  # no computation, just reshaping, so ignore

    def _add_gaussian_noise_layer(self, layer):
        return None  # no noise during testing


def LSUVinit(kmodel, X, tol=0.1, t_max=50):
    """Layer-sequential unit-variance initialization.

    References
    ----------
    .. [1] Mishkin, D., & Matas, J. (2016). All you need is a good init.
       In ICLR 2016 (pp. 1-13).
    """
    from keras.layers import Convolution2D, LocallyConnected2D
    import keras.backend as K
    # f = K.function([kmodel.layers[0].input, K.learning_phase()],
    #                [klayer.output])

    # --- orthogonalize weights
    def orthogonalize(X):
        assert X.ndim == 2
        U, s, V = np.linalg.svd(X, full_matrices=False)
        return np.dot(U, V)

    for layer in kmodel.layers:
        weights = layer.get_weights()
        if len(weights) == 0:
            continue

        W, b = weights
        if isinstance(layer, Convolution2D):
            Wv = W.reshape(W.shape[0], -1)
        elif isinstance(layer, LocallyConnected2D):
            Wv = W.reshape(-1, W.shape[-1])
        else:
            assert W.ndim == 2
            Wv = W

        Wv[:] = orthogonalize(Wv)
        layer.set_weights((W, b))

    # --- adjust variances
    s_input = kmodel.layers[0].input
    for layer in kmodel.layers:
        weights = layer.get_weights()
        if len(weights) == 0:
            continue

        W, b = weights
        f = K.function([s_input, K.learning_phase()], [layer.output])
        learning_phase = 0  # 0 == testing, 1 == training

        for i in range(t_max):
            Y = f([X, learning_phase])[0]
            Ystd = Y.std()
            print(Ystd)
            if abs(Ystd - 1) < tol:
                break

            W /= Ystd
            layer.set_weights((W, b))
        else:
            print("Layer %r did not converge after %d iterations (Ystd=%0.3e)"
                  % (layer.name, t_max, Ystd))



from __future__ import absolute_import

import collections
import os

import nengo
import numpy as np

from .compat import cmp, pickle_load
from .deepnetworks import SequentialNetwork


def load_model_pickle(loadfile):
    loadfile = os.path.expanduser(loadfile)
    with open(loadfile, 'rb') as f:
        return pickle_load(f)


def layer_depths(layers):
    depths = {}

    def get_depth(name):
        if name in depths:
            return depths[name]

        inputs = layers[name].get('inputs', [])
        depth = max(get_depth(i) for i in inputs) + 1 if len(inputs) > 0 else 0
        depths[name] = depth
        return depth

    for name in layers:
        get_depth(name)

    return depths


def sort_layers(layers, depths=None, cycle_check=True):
    depths = layer_depths(layers) if depths is None else depths

    def compare(a, b):
        da, db = depths[a], depths[b]
        return cmp(a, b) if da == db else cmp(da, db)

    snames = sorted(layers, cmp=compare)

    if cycle_check:
        def compare_cycle(a, b):
            ainb = a in layers[b].get('inputs', [])
            bina = b in layers[a].get('inputs', [])
            assert not (ainb and bina), "Cycle in graph"
            return -1 if ainb else 1 if bina else 0

        if any(compare_cycle(snames[i], snames[j]) < 0
               for i in range(1, len(snames)) for j in range(i)):
            raise ValueError("Cycle in graph")

    slayers = collections.OrderedDict((name, layers[name]) for name in snames)
    return slayers


class CudaConvnetNetwork(SequentialNetwork):
    def __init__(self, model, synapse=None, lif_type='lif', **kwargs):
        super(CudaConvnetNetwork, self).__init__(**kwargs)

        self.model = model
        self.synapse = synapse
        self.lif_type = lif_type

        # --- build model
        layers = dict(model['model_state']['layers'])

        # remove cost layer(s) and label inputs
        for name, layer in list(layers.items()):
            if layer['type'].startswith('cost'):
                layers.pop(name)
                layers.pop(layer['inputs'][0])  # first input is labels

        for layer in layers.values():
            assert all(i in layers for i in layer.get('inputs', ()))

        # sort layers
        depths = layer_depths(layers)
        assert np.unique(list(depths.values())).size == len(depths)

        for name in sorted(layers, key=lambda n: depths[n]):
            self._add_layer(layers[name])

    def _add_layer(self, layer):
        attrname = '_add_%s_layer' % layer['type'].replace('.', '_')
        if hasattr(self, attrname):
            return getattr(self, attrname)(layer)
        else:
            raise NotImplementedError(
                "Layer type %r not implemented" % layer['type'])

    def _get_inputs(self, layer):
        return [self.layers_by_name[i] for i in layer.get('inputs', [])]

    def _get_input(self, layer):
        assert len(layer.get('inputs', [])) == 1
        return self._get_inputs(layer)[0]

    def _add_data_layer(self, layer):
        d = layer['outputs']
        return self.add_data_layer(d, name=layer['name'])

    def _add_neuron_layer(self, layer):
        inputs = [self._get_input(layer)]
        neuron = layer['neuron']
        ntype = neuron['type']
        n = layer['outputs']

        gain = 1.
        bias = 0.
        amplitude = 1.
        if ntype == 'ident':
            neuron_type = nengo.Direct()
        elif ntype == 'relu':
            neuron_type = nengo.RectifiedLinear()
        elif ntype == 'logistic':
            neuron_type = nengo.Sigmoid()
        elif ntype == 'softlif':
            from .neurons import SoftLIFRate
            tau_ref, tau_rc, alpha, amp, sigma, noise = [
                neuron['params'][k] for k in ['t', 'r', 'a', 'm', 'g', 'n']]
            lif_type = self.lif_type.lower()
            if lif_type == 'lif':
                neuron_type = nengo.LIF(tau_rc=tau_rc, tau_ref=tau_ref)
            elif lif_type == 'lifrate':
                neuron_type = nengo.LIFRate(tau_rc=tau_rc, tau_ref=tau_ref)
            elif lif_type == 'softlifrate':
                neuron_type = SoftLIFRate(
                    sigma=sigma, tau_rc=tau_rc, tau_ref=tau_ref)
            else:
                raise KeyError("Unrecognized LIF type %r" % self.lif_type)
            gain = alpha
            bias = 1.
            amplitude = amp
        else:
            raise NotImplementedError("Neuron type %r" % ntype)

        return self.add_neuron_layer(
            n, inputs=inputs, neuron_type=neuron_type, synapse=self.synapse,
            gain=gain, bias=bias, amplitude=amplitude, name=layer['name'])

    def _add_softmax_layer(self, layer):
        return None  # non-neural, we can do without it
        # inputs = [self._get_input(layer)]
        # return self.add_softmax_layer(
        #     d=layer['outputs'], inputs=inputs, name=layer['name'])

    def _add_dropout_layer(self, layer):
        inputs = [self._get_input(layer)]
        d = layer['outputs']
        keep = layer['keep']
        return self.add_dropout_layer(
            d, keep, inputs=inputs, name=layer['name'])

    def _add_dropout2_layer(self, layer):
        return self._add_dropout_layer(layer)

    def _add_fc_layer(self, layer):
        inputs = [self._get_input(layer)]
        weights = layer['weights'][0].T
        biases = layer['biases'].ravel()
        return self.add_full_layer(
            weights, biases, inputs=inputs, name=layer['name'])

    def _add_conv_layer(self, layer):
        inputs = [self._get_input(layer)]
        assert layer['sharedBiases']
        nc = layer['channels'][0]
        nx = layer['imgSize'][0]
        ny = layer['modulesX']
        nf = layer['filters']
        s = layer['filterSize'][0]
        st = layer['stride'][0]
        p = -layer['padding'][0]

        filters = layer['weights'][0].reshape(nc, s, s, nf)
        filters = np.rollaxis(filters, axis=-1, start=0)
        biases = layer['biases']
        layer = self.add_conv_layer(
            (nc, nx, nx), filters, biases, strides=st, padding=p,
            border='ceil', inputs=inputs, name=layer['name'])
        assert layer.node.output.shape_out == (nf, ny, ny)
        return layer

    def _add_local_layer(self, layer):
        inputs = [self._get_input(layer)]
        nc = layer['channels'][0]
        nx = layer['imgSize'][0]
        ny = layer['modulesX']
        nf = layer['filters']
        s = layer['filterSize'][0]
        st = layer['stride'][0]
        p = -layer['padding'][0]

        filters = layer['weights'][0].reshape(ny, ny, nc, s, s, nf)
        filters = np.rollaxis(filters, axis=-1, start=0)
        biases = layer['biases'].reshape(nf, ny, ny)
        return self.add_local_layer(
            (nc, nx, nx), filters, biases, strides=st, padding=p,
            inputs=inputs, name=layer['name'])

    def _add_pool_layer(self, layer):
        inputs = [self._get_input(layer)]
        assert layer['start'] == 0
        nc = layer['channels']
        nx = layer['imgSize']
        s = layer['sizeX']
        st = layer['stride']
        kind = layer['pool']
        return self.add_pool_layer(
            (nc, nx, nx), s, strides=st, kind=kind, mode='full',
            inputs=inputs, name=layer['name'])



import re
import xml.etree.ElementTree as et

import nengo
import pytest

from nengo_extras.gexf import (
    CollapsingGexfConverter, DispatchTable, GexfConverter, HierarchicalLabeler)


def test_can_dispatch_table_defaults():
    class A(object):
        def __init__(self):
            self.called = False

    class Test(object):
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = True

    a = A()
    Test().dispatch(a)
    assert a.called


def test_dispatch_obj_inheritance():
    class A(object):
        def __init__(self):
            self.called = False

    class B(A):
        pass

    class C(A):
        pass

    class Test(object):
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = 'a'

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = 'b'

    a = A()
    Test().dispatch(a)
    assert a.called == 'a'

    b = B()
    Test().dispatch(b)
    assert b.called == 'b'

    c = C()
    Test().dispatch(c)
    assert c.called == 'a'


def test_dispatch_cls_inheritance():
    class A(object):
        def __init__(self):
            self.called = False

    class B(object):
        def __init__(self):
            self.called = False

    class C(object):
        def __init__(self):
            self.called = False

    class Test1(object):
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = 'test1.process_a'

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = 'test1.process_b'

    class Test2(Test1):
        dispatch = DispatchTable(Test1.dispatch)

        @dispatch.register(B)
        def process_b(self, obj):
            obj.called = 'test2.process_b'

        @dispatch.register(C)
        def process_c(self, obj):
            obj.called = 'test2.process_c'

    a = A()
    b = B()
    Test1().dispatch(a)
    assert a.called == 'test1.process_a'
    Test1().dispatch(b)
    assert b.called == 'test1.process_b'

    a = A()
    b = B()
    c = C()
    Test2().dispatch(a)
    assert a.called == 'test1.process_a'
    Test2().dispatch(b)
    assert b.called == 'test2.process_b'
    Test2().dispatch(c)
    assert c.called == 'test2.process_c'


def test_dispatch_instance_specific():
    class A(object):
        def __init__(self):
            self.called = False

    class B(object):
        def __init__(self):
            self.called = False

    class Test(object):
        dispatch = DispatchTable()

        @dispatch.register(A)
        def process_a(self, obj):
            obj.called = 'test.process_a'

    def proc_inst(obj):
        obj.called = 'proc_inst'

    test = Test()
    test2 = Test()
    test.dispatch.register(A, proc_inst)
    test.dispatch.register(B, proc_inst)

    a = A()
    test.dispatch(a)
    assert a.called == 'proc_inst'
    test2.dispatch(a)
    assert a.called == 'test.process_a'

    b = B()
    test.dispatch(b)
    assert b.called == 'proc_inst'


def test_dispatch_errors():
    class Test(object):
        dispatch = DispatchTable()

    with pytest.raises(NotImplementedError):
        Test().dispatch(object())


def test_hierarchical_labeler():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.)
            model.subnet.attr = nengo.Ensemble(10, 1)

    labels = HierarchicalLabeler().get_labels(model)
    assert dict(labels) == {
        model.ens: 'ens',
        model.subnet: 'subnet',
        model.subnet.node: 'subnet.node',
        model.subnet.attr: 'subnet.attr',
    }


def test_gexf_converter():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.)
            model.subnet.attr = nengo.Ensemble(10, 1)
        nengo.Connection(model.subnet.node, model.ens)
        nengo.Probe(model.ens)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        '<creator>nengo_extras.gexf.GexfConverter</creator>'
        '</meta>'
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        '<default>0</default>'
        '</attribute>'
        '<attribute id="7" title="neuron_type" type="string" />'
        '</attributes>'
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        '<default>1.0</default>'
        '</attribute>'
        '<attribute id="7" title="learning_rule_type" type="string" />'
        '</attributes>'
        '<nodes>'
        '<node id="\\d+" label="ens">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        '</attvalues>'
        '</node>'
        '<node id="\\d+" label="subnet.attr">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        '</attvalues>'
        '</node>'
        '<node id="\\d+" label="subnet.node">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="0" />'
        '<attvalue for="4" value="1" />'
        '</attvalues>'
        '</node>'
        '</nodes>'
        '<edges>'
        '<edge id="\\d+" source="\\d+" target="\\d+">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="2" value="Lowpass\\(0.005\\)" />'
        '<attvalue for="3" value="0.005" />'
        '<attvalue for="5" value="1.0" />'
        '</attvalues>'
        '</edge>'
        '</edges>'
        '</graph>'
        '</gexf>$')

    actual = et.tostring(
        GexfConverter().convert(model).getroot()).decode("utf-8")
    assert re.match(expected, actual), actual


def test_gexf_converter_hierarchical():
    with nengo.Network() as model:
        model.ens = nengo.Ensemble(10, 1)
        with nengo.Network() as model.subnet:
            model.subnet.node = nengo.Node(1.)
            model.subnet.attr = nengo.Ensemble(10, 1)
        nengo.Connection(model.subnet.node, model.ens)
        nengo.Probe(model.ens)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        '<creator>nengo_extras.gexf.GexfConverter</creator>'
        '</meta>'
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        '<default>0</default>'
        '</attribute>'
        '<attribute id="7" title="neuron_type" type="string" />'
        '</attributes>'
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        '<default>1.0</default>'
        '</attribute>'
        '<attribute id="7" title="learning_rule_type" type="string" />'
        '</attributes>'
        '<nodes>'
        '<node id="\\d+" label="ens">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        '</attvalues>'
        '</node>'
        '<node id="\\d+" label="subnet">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.network.Network" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="6" value="10" />'
        '</attvalues>'
        '<nodes>'
        '<node id="\\d+" label="subnet.attr">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="1" />'
        '<attvalue for="4" value="1" />'
        '<attvalue for="5" value="1.0" />'
        '<attvalue for="6" value="10" />'
        '<attvalue for="7" value="LIF\\(\\)" />'
        '</attvalues>'
        '</node>'
        '<node id="\\d+" label="subnet.node">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="subnet" />'
        '<attvalue for="3" value="0" />'
        '<attvalue for="4" value="1" />'
        '</attvalues>'
        '</node>'
        '</nodes>'
        '</node>'
        '</nodes>'
        '<edges>'
        '<edge id="\\d+" source="\\d+" target="\\d+">'
        '<attvalues>'
        '<attvalue for="0" value="nengo.node.Node" />'
        '<attvalue for="1" value="nengo.ensemble.Ensemble" />'
        '<attvalue for="2" value="Lowpass\\(0.005\\)" />'
        '<attvalue for="3" value="0.005" />'
        '<attvalue for="5" value="1.0" />'
        '</attvalues>'
        '</edge>'
        '</edges>'
        '</graph>'
        '</gexf>$')

    actual = et.tostring(GexfConverter(
        hierarchical=True).convert(model).getroot()).decode("utf-8")
    assert re.match(expected, actual), actual


def test_collapsing_gexf_converter():
    with nengo.Network() as model:
        model.ea = nengo.networks.EnsembleArray(10, 10)

    expected = (
        '<gexf version="1.3" '
        'xmlns="http://www.gexf.net/1.3draft" '
        'xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance" '
        'xsi:schemaLocation="http://www.gexf.net/1.3draft '
        'http://www.gexf.net/1.3draft/gexf.xsd">'
        '<meta lastmodifieddate="\\d+-\\d{2}-\\d{2}">'
        '<creator>nengo_extras.gexf.CollapsingGexfConverter</creator>'
        '</meta>'
        '<graph defaultedgetype="directed">'
        '<attributes class="node">'
        '<attribute id="0" title="type" type="string" />'
        '<attribute id="1" title="net" type="long" />'
        '<attribute id="2" title="net_label" type="string" />'
        '<attribute id="3" title="size_in" type="integer" />'
        '<attribute id="4" title="size_out" type="integer" />'
        '<attribute id="5" title="radius" type="float" />'
        '<attribute id="6" title="n_neurons" type="integer">'
        '<default>0</default>'
        '</attribute>'
        '<attribute id="7" title="neuron_type" type="string" />'
        '</attributes>'
        '<attributes class="edge">'
        '<attribute id="0" title="pre_type" type="string" />'
        '<attribute id="1" title="post_type" type="string" />'
        '<attribute id="2" title="synapse" type="string" />'
        '<attribute id="3" title="tau" type="float" />'
        '<attribute id="4" title="function" type="string" />'
        '<attribute id="5" title="transform" type="string" />'
        '<attribute id="6" title="scalar_transform" type="float">'
        '<default>1.0</default>'
        '</attribute>'
        '<attribute id="7" title="learning_rule_type" type="string" />'
        '</attributes>'
        '<nodes>'
        '<node id="\\d+" label="ea">'
        '<attvalues>'
        '<attvalue for="0" value="'
        'nengo.networks.ensemblearray.EnsembleArray" />'
        '<attvalue for="1" value="\\d+" />'
        '<attvalue for="2" value="model" />'
        '<attvalue for="6" value="100" />'
        '</attvalues>'
        '</node>'
        '</nodes>'
        '<edges>'
        '.*'
        '</edges>'
        '</graph>'
        '</gexf>$')

    actual = et.tostring(
        CollapsingGexfConverter().convert(model).getroot()).decode("utf-8")
    assert re.match(expected, actual), actual



import numpy as np

from nengo.dists import Choice, Uniform, DistributionParam
from nengo.params import FrozenObject, TupleParam
from nengo.utils.compat import range
import nengo.utils.numpy as npext


class Gabor(FrozenObject):
    """Describes a random generator for Gabor filters."""

    theta = DistributionParam('theta')
    freq = DistributionParam('freq')
    phase = DistributionParam('phase')
    sigma_x = DistributionParam('sigma_x')
    sigma_y = DistributionParam('sigma_y')

    def __init__(self, theta=Uniform(-np.pi, np.pi), freq=Uniform(0.2, 2),
                 phase=Uniform(-np.pi, np.pi),
                 sigma_x=Choice([0.45]), sigma_y=Choice([0.45])):
        super(Gabor, self).__init__()
        self.theta = theta
        self.freq = freq
        self.phase = phase
        self.sigma_x = sigma_x
        self.sigma_y = sigma_y

    def generate(self, n, shape, rng=np.random, norm=1.):
        assert isinstance(shape, tuple) and len(shape) == 2
        thetas = self.theta.sample(n, rng=rng)[:, None, None]
        freqs = self.freq.sample(n, rng=rng)[:, None, None]
        phases = self.phase.sample(n, rng=rng)[:, None, None]
        sigma_xs = self.sigma_x.sample(n, rng=rng)[:, None, None]
        sigma_ys = self.sigma_y.sample(n, rng=rng)[:, None, None]

        x, y = np.linspace(-1, 1, shape[1]), np.linspace(-1, 1, shape[0])
        X, Y = np.meshgrid(x, y)

        c, s = np.cos(thetas), np.sin(thetas)
        X1 = X * c + Y * s
        Y1 = -X * s + Y * c

        gabors = np.exp(-0.5 * ((X1 / sigma_xs)**2 + (Y1 / sigma_ys)**2))
        gabors *= np.cos((2 * np.pi) * freqs * X1 + phases)

        if norm is not None:
            gabors *= norm / np.sqrt(
                (gabors**2).sum(axis=(1, 2), keepdims=True)).clip(1e-5, np.inf)

        return gabors


class Mask(FrozenObject):
    """Describes a sparse receptive-field mask for encoders.

    Parameters
    ----------
    image_shape : 2- or 3-tuple
        Shape of the input image, either (height, witdh) or
        (channels, height, width).
    """

    image_shape = TupleParam('image_shape', length=3)

    def __init__(self, image_shape):
        super(Mask, self).__init__()
        image_shape = ((1,) + tuple(image_shape) if len(image_shape) == 2 else
                       image_shape)
        self.image_shape = image_shape

    def _positions(self, n, shape, rng):
        diff_shape = np.asarray(self.image_shape[1:]) - np.asarray(shape) + 1

        # find random positions for top-left corner of each RF
        i = rng.randint(low=0, high=diff_shape[0], size=n)
        j = rng.randint(low=0, high=diff_shape[1], size=n)
        return i, j

    def generate(self, n, shape, rng=np.random, flatten=False):
        shape = np.asarray(shape)
        assert shape.ndim == 1 and shape.shape[0] == 2

        i, j = self._positions(n, shape, rng)
        mask = np.zeros((n,) + self.image_shape, dtype='bool')
        for k in range(n):
            mask[k, :, i[k]:i[k]+shape[0], j[k]:j[k]+shape[1]] = True

        return mask.reshape(n, -1) if flatten else mask

    def populate(self, filters, rng=np.random, flatten=False):
        filters = np.asarray(filters)
        assert filters.ndim in [3, 4]
        n, shape = filters.shape[0], filters.shape[-2:]
        channels = 1 if filters.ndim == 3 else filters.shape[1]
        assert channels == self.image_shape[0]

        i, j = self._positions(n, shape, rng)
        output = np.zeros((n,) + self.image_shape, dtype=filters.dtype)
        for k in range(n):
            output[k, :, i[k]:i[k]+shape[0], j[k]:j[k]+shape[1]] = filters[k]

        return output.reshape(n, -1) if flatten else output


def ciw_encoders(n_encoders, trainX, trainY, rng=np.random,
                 normalize_data=True, normalize_encoders=True):
    """Computed Input Weights (CIW) method for encoders from data.

    Parameters
    ----------
    n_encoders : int
        Number of encoders to generate.
    trainX : (n_samples, n_dimensions) array-like
        Training features.
    trainY : (n_samples,) array-like
        Training labels.

    Returns
    -------
    encoders : (n_encoders, n_dimensions) array
        Generated encoders.

    References
    ----------
    .. [1] McDonnell, M. D., Tissera, M. D., Vladusich, T., Van Schaik, A.,
       Tapson, J., & Schwenker, F. (2015). Fast, simple and accurate
       handwritten digit classification by training shallow neural network
       classifiers with the "Extreme learning machine" algorithm. PLoS ONE,
       10(8), 1-20. doi:10.1371/journal.pone.0134254
    """
    assert trainX.shape[0] == trainY.size
    trainX = trainX.reshape(trainX.shape[0], -1)
    trainY = trainY.ravel()
    classes = np.unique(trainY)

    assert n_encoders % len(classes) == 0
    n_enc_per_class = n_encoders / len(classes)

    # normalize
    if normalize_data:
        trainX = (trainX - trainX.mean()) / trainX.std()
        # trainX = (trainX - trainX.mean(axis=0)) / trainX.std()
        # trainX = (trainX - trainX.mean(axis=0)) / (trainX.std(axis=0) + 1e-8)

    # generate
    encoders = []
    for label in classes:
        X = trainX[trainY == label]
        plusminus = rng.choice([-1, 1], size=(X.shape[0], n_enc_per_class))
        samples = np.dot(plusminus.T, X)
        encoders.append(samples)

    encoders = np.vstack(encoders)
    if normalize_encoders:
        encoders /= npext.norm(encoders, axis=1, keepdims=True)

    return encoders


def cd_encoders_biases(n_encoders, trainX, trainY, rng=np.random, mask=None,
                       norm_min=0.05, norm_tries=10):
    """Constrained difference (CD) method for encoders from data.

    Parameters
    ----------
    n_encoders : int
        Number of encoders to generate.
    trainX : (n_samples, n_dimensions) array-like
        Training features.
    trainY : (n_samples,) array-like
        Training labels.

    Returns
    -------
    encoders : (n_encoders, n_dimensions) array
        Generated encoders.
    biases : (n_encoders,) array
        Generated biases. These are biases assuming ``f = G[E * X + b]``,
        and are therefore more like Nengo's ``intercepts``.

    References
    ----------
    .. [1] McDonnell, M. D., Tissera, M. D., Vladusich, T., Van Schaik, A.,
       Tapson, J., & Schwenker, F. (2015). Fast, simple and accurate
       handwritten digit classification by training shallow neural network
       classifiers with the "Extreme learning machine" algorithm. PLoS ONE,
       10(8), 1-20. doi:10.1371/journal.pone.0134254
    """
    assert trainX.shape[0] == trainY.size
    trainX = trainX.reshape(trainX.shape[0], -1)
    trainY = trainY.ravel()
    d = trainX.shape[1]
    classes = np.unique(trainY)
    assert mask is None or mask.shape == (n_encoders, d)

    inds = [(trainY == label).nonzero()[0] for label in classes]
    train_norm = npext.norm(trainX, axis=1).mean()

    encoders = np.zeros((n_encoders, d))
    biases = np.zeros(n_encoders)
    for k in range(n_encoders):
        for _ in range(norm_tries):
            i, j = rng.choice(len(classes), size=2, replace=False)
            a, b = trainX[rng.choice(inds[i])], trainX[rng.choice(inds[j])]
            dab = a - b
            if mask is not None:
                dab *= mask[k]
            ndab = npext.norm(dab)**2
            if ndab >= norm_min * train_norm:
                break
        else:
            raise ValueError("Cannot find valid encoder")

        encoders[k] = (2. / ndab) * dab
        biases[k] = np.dot(a + b, dab) / ndab

    return encoders, biases


def percentile_biases(encoders, trainX, percentile=50):
    """Pick biases such that neurons are active for a percentile of inputs."""
    trainX = trainX.reshape(trainX.shape[0], -1)
    H = np.dot(trainX, encoders.T)
    biases = np.percentile(H, percentile, axis=0)
    return biases



import pytest

import nengo

from nengo_extras.probe import probe_all


def make_net():
    with nengo.Network() as net:
        net.ens1 = nengo.Ensemble(n_neurons=1, dimensions=1)
        net.node1 = nengo.Node(output=lambda t, x: x, size_in=1)
        net.conn = nengo.Connection(net.ens1, net.node1,
                                    learning_rule_type=[nengo.PES()])

        with nengo.Network() as net.subnet:
            net.ens2 = nengo.Ensemble(n_neurons=1, dimensions=1)
            net.node2 = nengo.Node(output=[0])

    return net


@pytest.mark.parametrize("recursive", [False, True])
def test_probe_all_recursive(recursive):
    net = make_net()
    probes = probe_all(net, recursive=recursive)

    assert(len(probes[net.ens1]) == len(net.ens1.probeable))
    assert(len(probes[net.ens1.neurons]) == len(net.ens1.neurons.probeable))
    assert(len(probes[net.node1]) == len(net.node1.probeable))
    assert(len(probes[net.conn]) == len(net.conn.probeable))
    for lr in net.conn.learning_rule:
        assert(len(probes[lr]) == len(lr.probeable))

    if recursive:
        assert(len(probes) == 8)
        assert(len(probes[net.ens2]) == len(net.ens2.probeable))
        assert(len(probes[net.node2]) == len(net.node2.probeable))
    else:
        assert(len(probes) == 5)
        assert net.ens2 not in probes
        assert net.ens2.neurons not in probes
        assert net.node2 not in probes


@pytest.mark.parametrize("recursive", [False, True])
def test_probe_all_options(recursive):
    net = make_net()

    probe_all(net, recursive=recursive, probe_options={
        nengo.Ensemble: ['decoded_output']})

    if recursive:
        assert(len(net.probes) == 2)
    else:
        assert(len(net.probes) == 1)


def test_probe_all_kwargs():
    net = make_net()
    probe_all(net, recursive=True, sample_every=0.1, seed=10)

    for probe in net.probes:
        assert probe.sample_every == 0.1
        assert probe.seed == 10
